"""
First file to be imported when import wdoc by __main__
It's purpose is to detect wether we should use import tricks like
lazyloading or threaded imports
"""

import sys

from .env import WDOC_IMPORT_TYPE


def trick_imports() -> None:
    import threading
    from queue import Queue

    import lazy_import

    def import_worker(q: Queue):
        while True:
            module = q.get()
            if module is None:
                return
            # print(f"Importing {module}")
            if "." in module:
                first = ".".join(module.split(".")[:-1])
                last = module.split(".")[-1]
                exec(f"from {first} import {last}")
            else:
                first = module
                exec(f"import {module}")

            assert first in sys.modules, f"Error when importing '{first}'"
            if "Lazily-loaded" in str(sys.modules[first]):
                # print(f"Module is lazy loaded so far: {first}")
                try:
                    dir(sys.modules[first])
                except Exception as e:
                    print(
                        f"Error when unlazyloading module '{first}'. Error: '{e}'"
                        "\nYou can try setting the env variable "
                        "WDOC_DISABLE_LAZYLOADING to 'true'"
                        "Don't hesitate to open an issue!"
                    )
                # print(f"Unlazyloaded module: {first}")

    if WDOC_IMPORT_TYPE == "thread":
        q = Queue()
        thread = threading.Thread(
            target=import_worker,
            args=(q,),
            daemon=False,
        )
        thread.start()

    def custom_loading(module: str, strict: bool = True) -> None:
        """
        Bunch of import tricks. The strict parameter allows to specify module
        that can be tricky to install or import.
        """
        if not strict:
            try:
                custom_loading(module=module)
            except Exception:
                return
        if WDOC_IMPORT_TYPE == "both":
            lazy_import.lazy_module(module)
            q.put(module)
        elif WDOC_IMPORT_TYPE == "thread":
            q.put(module)
        elif WDOC_IMPORT_TYPE == "lazy":
            if "module" in sys.modules:
                print(f"{module} already imported")
            lazy_import.lazy_module(module)
            assert module in sys.modules, module
        else:
            raise ValueError(
                f"Unexpected value for WDOC_IMPORT_TYPE: '{WDOC_IMPORT_TYPE}'"
            )

    custom_loading("langchain")
    custom_loading("langchain_community")
    custom_loading("langchain.text_splitter")
    custom_loading("langfuse")
    custom_loading("litellm")
    custom_loading("numpy")
    custom_loading("faiss")
    custom_loading("zlib")
    custom_loading("dill")
    custom_loading("sqlite3")
    custom_loading("tldextract")
    custom_loading("pyfiglet")
    custom_loading("yt_dlp", strict=False)
    custom_loading("pandas")
    custom_loading("ankipandas")
    custom_loading("ftfy")
    custom_loading("bs4")
    custom_loading("goose3")
    custom_loading("LogseqMarkdownParser")
    custom_loading("deepgram")
    custom_loading("pydub")
    custom_loading("ffmpeg")
    custom_loading("torchaudio", strict=False)
    custom_loading("playwright.sync_api")
    custom_loading("openparse")
    custom_loading("scipy")
    custom_loading("sklearn.metrics")
    custom_loading("sklearn.decomposition")
    custom_loading("sklearn.preprocessing")

    if WDOC_IMPORT_TYPE in ["both", "thread"]:
        q.put(None)  # kill the import worker


if WDOC_IMPORT_TYPE != "native":
    trick_imports()



"""
source : https://api.python.langchain.com/en/latest/_modules/langchain_community/cache.html#InMemoryCache

This workaround is to solve this: https://github.com/langchain-ai/langchain/issues/22389
Create a caching class that looks like it's just in memory but actually saves to sql

"""

import json
from pathlib import Path

from beartype.typing import Any, Generator, Optional, Union
from langchain_core.caches import BaseCache
from PersistDict import PersistDict


class SQLiteCacheFixed(BaseCache):
    """Cache that stores things in memory using SQLiteDict."""

    def __init__(
        self,
        database_path: Union[str, Path],
        expiration_days: Optional[int] = 0,
        verbose: bool = False,
    ) -> None:
        self.pdi = PersistDict(
            database_path=database_path,
            expiration_days=expiration_days,
            verbose=verbose,
        )

    def lookup(self, prompt: str, llm_string: str) -> Any:
        """Look up based on prompt and llm_string."""
        key = json.dumps((prompt, llm_string))
        try:
            val = self.pdi[key]
        except KeyError:
            return None
        return val

    def update(self, prompt: str, llm_string: str, return_val: Any) -> None:
        """Update cache based on prompt and llm_string."""
        key = json.dumps((prompt, llm_string))
        self.pdi[key] = return_val

    def clear(self) -> None:
        raise NotImplementedError()
        # self.pdi.clear()

    async def alookup(self, prompt: str, llm_string: str) -> Any:
        """Look up based on prompt and llm_string."""
        return self.lookup(prompt, llm_string)

    async def aupdate(self, prompt: str, llm_string: str, return_val: Any) -> None:
        """Update cache based on prompt and llm_string."""
        self.update(prompt, llm_string, return_val)

    async def aclear(self) -> None:
        """Clear cache."""
        self.clear()

    def __get_keys__(self) -> Generator[str, None, None]:
        "get the list of keys present in the db"
        for k in self.pdi.keys():
            yield k



"""
Code related to loggings, coloured logs, etc.
"""

import json
import warnings
from pathlib import Path
from textwrap import dedent

import rtoml
from beartype.typing import Callable, Dict, List, Optional, Type, Union
from loguru import logger
from platformdirs import user_cache_dir, user_log_dir
from rich.console import Console
from rich.markdown import Markdown
from tqdm import tqdm

from .flags import is_silent, md_printing_disabled
from .typechecker import optional_typecheck

# ignore warnings from beautiful soup
warnings.filterwarnings("ignore", category=UserWarning, module="bs4")

cache_dir = Path(user_cache_dir(appname="wdoc"))
cache_dir.mkdir(parents=True, exist_ok=True)

log_dir = Path(user_log_dir(appname="wdoc"))
log_dir.mkdir(exist_ok=True, parents=True)
log_file = log_dir / "logs.txt"
log_file.touch(exist_ok=True)

# logger
logger.add(
    log_file,
    rotation="100MB",
    retention=5,
    format="{time} {level} wdoc {thread} {process} {function} {line} {message}",
    level="DEBUG",
    enqueue=False,
    colorize=False,
)
# delete any additional log file
# (log_dir / "logs.txt.4").unlink(missing_ok=True)


colors = {
    "red": "\033[91m",
    "yellow": "\033[93m",
    "reset": "\033[0m",
    "white": "\033[0m",
    "purple": "\033[95m",
}


@optional_typecheck
def get_coloured_logger(color_asked: str) -> Callable:
    """used to print color coded logs"""
    col = colors[color_asked]

    # all logs are considered "errors" otherwise the datascience libs just
    # overwhelm the logs
    @optional_typecheck
    def printer(string: Union[str, Dict, List, Exception], **args) -> str:
        if isinstance(string, Exception):
            string = str(string)
        if isinstance(string, dict):
            try:
                string = rtoml.dumps(string, pretty=True)
            except Exception:
                string = json.dumps(string, indent=2, ensure_ascii=False)
        if isinstance(string, list):
            try:
                string = ",".join(string)
            except Exception:
                pass
        try:
            string = str(string)
        except Exception:
            try:
                string = string.__str__()
            except Exception:
                string = string.__repr__()
        for k, v in colors.items():
            string = string.replace(v, "")
        logger.info(string)
        if not is_silent:
            tqdm.write(col + string + colors["reset"], **args)
        return string

    return printer


whi = get_coloured_logger("white")
yel = get_coloured_logger("yellow")
red = get_coloured_logger("red")

console = Console()


@optional_typecheck
def md_printer(message: str, color: Optional[str] = None) -> str:
    "markdown printing"
    message = dedent(message)
    if not md_printing_disabled:
        logger.info(message)
        md = Markdown(message)
        console.print(md, style=color)
    else:
        if not color:
            whi(message)
        elif color in "red":
            red(message)
        elif color in "white":
            whi(message)
        elif color in "yellow":
            yel(message)
        else:
            whi(message)
    return message


@optional_typecheck
def set_help_md_as_docstring(obj: Union[Type, Callable]) -> Union[Type, Callable]:
    "set the docstring of wdoc class to wdoc/docs/help.md's content"
    help_file = Path(__file__).parent.parent / "docs/help.md"
    if not help_file.exists():
        red(
            f"Couldn't find help.md file as '{help_file}'. You can read it at this URL instead: https://github.com/thiswillbeyourgithub/wdoc/blob/main/wdoc/docs/help.md"
        )
        helpcont = "Help documentation not found. Please refer to online documentation."
    else:
        helpcont = help_file.read_text().strip()
        if not helpcont:
            red("Help documentation file is empty")
            helpcont = (
                "Help documentation is empty. Please refer to online documentation."
            )
    obj.__doc__ = "# Content of wdoc/docs/help.md\n\n" + helpcont
    if isinstance(obj, type):
        obj.__init__.__doc__ = helpcont
    return obj


@optional_typecheck
def set_parse_file_help_md_as_docstring(
    obj: Union[Type, Callable],
) -> Union[Type, Callable]:
    "set the docstring of wdoc.parse_file to wdoc/docs/parse_file_help.md's content"
    parsefilehelp_file = Path(__file__).parent.parent / "docs/parse_file_help.md"
    if not parsefilehelp_file.exists():
        red(
            f"Couldn't find parse_file_help.md file as '{parsefilehelp_file}'. You can read it at this URL instead: https://github.com/thiswillbeyourgithub/wdoc/blob/main/wdoc/docs/parse_file_help.md"
        )
        parsefilehelp = "Parse file help documentation not found. Please refer to online documentation."
    else:
        parsefilehelp = parsefilehelp_file.read_text().strip()
        if not parsefilehelp:
            red("Parse file help documentation is empty")
            parsefilehelp = "Parse file help documentation is empty. Please refer to online documentation."
    obj.__doc__ = "# Content of wdoc/docs/parse_file_help.md\n\n" + parsefilehelp
    return obj



"""
Chain (logic) used to summarize a document.
"""

from pathlib import Path

from beartype.typing import Any, List, Tuple, Union
from langchain.docstore.document import Document
from tqdm import tqdm

from ..logger import red, whi
from ..misc import thinking_answer_parser
from ..prompts import BASE_SUMMARY_PROMPT, PREV_SUMMARY_TEMPLATE, RECURSION_INSTRUCTION
from ..typechecker import optional_typecheck

HOME = str(Path.home())


@optional_typecheck
def do_summarize(
    docs: List[Document],
    metadata: str,
    language: str,
    modelbackend: str,
    llm: Any,
    llm_price: List[float],
    verbose: bool,
    n_recursion: int = 0,
) -> Tuple[str, int, Union[float, int], int, int]:
    "summarize each chunk of a long document"
    summaries = []
    previous_summary = ""

    llm.bind(verbose=verbose)

    total_tokens = [0, 0]
    total_cost = 0

    metadata = metadata.replace(HOME, "~")  # extra privacy just in case a path appears

    assert "[PROGRESS]" in metadata
    for ird, rd in tqdm(enumerate(docs), desc="Summarising splits", total=len(docs)):
        fixed_index = f"{ird + 1}/{len(docs)}"

        messages = BASE_SUMMARY_PROMPT.format_messages(
            language=language,
            metadata=metadata.replace("[PROGRESS]", fixed_index),
            previous_summary=previous_summary,
            recursion_instruction="" if not n_recursion else RECURSION_INSTRUCTION,
            text=rd.page_content,
        )
        if " object at " in llm._get_llm_string():
            red(
                "Found llm._get_llm_string() value that potentially "
                f"invalidates the cache: '{llm._get_llm_string()}'\n"
                f"Related github issue: 'https://github.com/langchain-ai/langchain/issues/23257'"
            )
        try:
            output = llm._generate_with_cache(messages)
        except Exception as e:
            red(f"Error when generating with cache, trying without cache: '{e}'")
            output = llm._generate(messages)
        if output.generations[0].generation_info is None:
            assert "fake-list-chat-model" in llm._get_llm_string()
            finish = "stop"
        else:
            finish = output.generations[0].generation_info["finish_reason"]
            assert finish == "stop", f"Unexpected finish_reason: '{finish}'"
            assert len(output.generations) == 1
        out = output.generations[0].text
        if output.llm_output:  # only present if not caching
            new_p = output.llm_output["token_usage"]["prompt_tokens"]
            new_c = output.llm_output["token_usage"]["completion_tokens"]
        else:
            new_p = 0
            new_c = 0
        total_tokens[0] += new_p
        total_tokens[1] += new_c
        total_cost += (new_p * llm_price[0] + new_c + llm_price[1]) / 1e6

        # the callback need to be updated manually when _generate is called
        llm.callbacks[0].prompt_tokens += new_p
        llm.callbacks[0].completion_tokens += new_c
        llm.callbacks[0].total_tokens += new_p + new_c

        parsed = thinking_answer_parser(out)
        if verbose and parsed["thinking"]:
            whi("Thinking: " + parsed["thinking"])

        output_lines = parsed["answer"].rstrip().splitlines(keepends=True)

        for il, ll in enumerate(output_lines):
            # remove if contains no alphanumeric character
            if not any(char.isalpha() for char in ll.strip()):
                output_lines[il] = None
                continue

            ll = ll.rstrip()

            # replace tabs by 4 spaces
            ll = ll.replace("\t", "    ")
            ll = ll.replace("	", "    ")

            stripped = ll.lstrip()

            ll = ll.replace("- • ", "- ").replace("• ", "- ")  # occasional bad md

            # if a line starts with * instead of -, fix it
            if stripped.startswith("* "):
                ll = ll.replace("*", "-", 1)

            stripped = ll.lstrip()
            # beginning with long dash
            if stripped.startswith("—"):
                ll = ll.replace("—", "-")

            # begin by '-' but not by '- '
            stripped = ll.lstrip()
            if stripped.startswith("-") and not stripped.startswith("- "):
                ll = ll.replace("-", "- ", 1)

            # if a line does not start with - fix it
            stripped = ll.lstrip()
            if not stripped.startswith("- "):
                ll = ll.replace(stripped[0], "- " + stripped[0], 1)

            ll = ll.replace("****", "")

            # if contains uneven number of bold markers
            if ll.count("**") % 2 == 1:
                ll += "**"  # end the bold
            # and italic
            if ll.count("*") % 2 == 1:
                ll += "*"  # end the italic

            output_lines[il] = ll

        good_lines = [li for li in output_lines if li]
        output_text = "\n".join(good_lines)

        if verbose:
            whi(output_text)

        previous_summary = PREV_SUMMARY_TEMPLATE.replace(
            "{previous_summary}",
            "...\n" + "\n".join(good_lines[-5:]),
        )

        summaries.append(output_text)

    # combine summaries as one string separated by markdown separator
    n = len(summaries)
    if n > 1:
        outtext = f"- Chunk 1/{n}\n"
        for i, s in enumerate(summaries):
            outtext += s + "\n"
            if n > 1 and s != summaries[-1]:
                outtext += f"- ---\n- Chunk {i + 2}/{n}\n"
    else:
        outtext = "\n".join(summaries)

    return outtext.rstrip(), n, total_cost, total_tokens[0], total_tokens[1]



"""
Decorator used in many places. It does runtime typechecking. By default
it's disabled but the flag WDOC_TYPECHECKING can be set to "crash" or
to "warn" to just print the error.
"""

from beartype import BeartypeConf, beartype
from beartype.typing import Callable

from .env import WDOC_TYPECHECKING

if WDOC_TYPECHECKING == "crash":
    optional_typecheck = beartype
elif WDOC_TYPECHECKING == "warn":
    optional_typecheck = beartype(conf=BeartypeConf(violation_type=UserWarning))
elif WDOC_TYPECHECKING == "disabled":

    def optional_typecheck(func: Callable) -> Callable:
        return func

else:
    raise ValueError("Unexpected WDOC_TYPECHECKING env value")



import tempfile
from pathlib import Path

import pytest


def pytest_configure(config):
    """Register custom markers."""
    config.addinivalue_line(
        "markers",
        "basic: mark test as a basic test that doesn't require external services",
    )
    config.addinivalue_line(
        "markers", "api: mark test as requiring external API access"
    )


@pytest.fixture
def temp_dir():
    """Provide a temporary directory that is cleaned up after the test."""
    with tempfile.TemporaryDirectory() as tmpdirname:
        yield Path(tmpdirname)


@pytest.fixture
def sample_text_file(temp_dir):
    """Create a sample text file for testing."""
    file_path = temp_dir / "sample.txt"
    with open(file_path, "w") as f:
        f.write(
            "This is a test document.\nIt has multiple lines.\nFor testing purposes."
        )
    return file_path


@pytest.fixture
def sample_pdf_file(temp_dir):
    """Create a sample PDF file path for testing."""
    return temp_dir / "sample.pdf"



"""
Miscellanous functions etc.
"""

import hashlib
import inspect
import json
import os
import socket
import sys
import uuid
import warnings
from dataclasses import dataclass, field
from datetime import timedelta
from difflib import get_close_matches
from functools import cache as memoize
from functools import partial, wraps
from pathlib import Path

import bs4
import litellm
from beartype.door import is_bearable
from beartype.typing import Callable, List, Literal, Union, get_type_hints
from joblib import Memory
from joblib import hash as jhash
from langchain.docstore.document import Document
from langchain.text_splitter import RecursiveCharacterTextSplitter, TextSplitter
from langchain_core.runnables import chain
from py_ankiconnect import PyAnkiconnect

from .env import (
    WDOC_ALLOW_NO_PRICE,
    WDOC_EXPIRE_CACHE_DAYS,
    WDOC_IMPORT_TYPE,
    WDOC_MAX_CHUNK_SIZE,
    WDOC_NO_MODELNAME_MATCHING,
    WDOC_PRIVATE_MODE,
    WDOC_STRICT_DOCDICT,
)
from .errors import UnexpectedDocDictArgument
from .flags import is_debug, is_private, is_verbose
from .logger import cache_dir, red, whi, yel
from .typechecker import optional_typecheck

ankiconnect = optional_typecheck(PyAnkiconnect())

# will be replaced when load_one_doc is called, by the path to the file where the loaders can store temporary file
loaders_temp_dir_file = cache_dir / "loaders_temp_dir.txt"

# ignore warnings from beautiful soup that can happen because anki is not exactly html
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    module="bs4",
    message=".*The input looks more like a filename than markup.*",
)

# additional warnings to ignore
warnings.filterwarnings(
    "ignore", module="litellm", message=".*Counting tokens for OpenAI model=.*"
)

try:
    import ftlangdetect

    @optional_typecheck
    def language_detector(text: str) -> float:
        return ftlangdetect.detect(text.lower())["score"]

    assert isinstance(language_detector("This is a test"), float)
except Exception as err:
    if is_verbose:
        red(
            f"Couldn't import optional package 'ftlangdetect' from 'fasttext-langdetect', trying to import langdetect (but it's much slower): '{err}'"
        )
    if "ftlangdetect" in sys.modules:
        del sys.modules["ftlangdetect"]

    try:
        import langdetect

        @optional_typecheck
        def language_detector(text: str) -> float:
            return langdetect.detect_langs(text.lower())[0].prob

        assert isinstance(language_detector("This is a test"), float)
    except Exception as err:
        if is_verbose:
            red(f"Couldn't import optional package 'langdetect' either: '{err}'")

        @optional_typecheck
        def language_detector(text: str) -> None:
            return None


doc_loaders_cache_dir = cache_dir / "doc_loaders"
doc_loaders_cache_dir.mkdir(exist_ok=True)
doc_loaders_cache = Memory(doc_loaders_cache_dir, verbose=0)
hashdoc_cache_dir = cache_dir / "doc_hashing"
hashdoc_cache_dir.mkdir(exist_ok=True)
hashdoc_cache = Memory(hashdoc_cache_dir, verbose=0)
(cache_dir / "query_eval_llm").mkdir(exist_ok=True)
query_eval_cache = Memory(cache_dir / "query_eval_llm", verbose=0)

# remove cache files older than X days
if WDOC_EXPIRE_CACHE_DAYS:
    doc_loaders_cache.reduce_size(age_limit=timedelta(days=int(WDOC_EXPIRE_CACHE_DAYS)))
    hashdoc_cache.reduce_size(age_limit=timedelta(days=int(WDOC_EXPIRE_CACHE_DAYS)))
    query_eval_cache.reduce_size(age_limit=timedelta(days=int(WDOC_EXPIRE_CACHE_DAYS)))

# for reading length estimation
wpm = 250
average_word_length = 6

# separators used for the text splitter
recur_separator = ["\n\n\n\n", "\n\n\n", "\n\n", "\n", "...", ".", " ", ""]

min_token = 20
max_token = 10_000_000
min_lang_prob = 0.50

printed_unexpected_api_keys = [False]  # to print it only once

# loader specific arguments
filetype_arg_types = {
    "pdf_parsers": Union[str, List[str]],
    "anki_deck": str,
    "anki_notetype": str,
    "anki_profile": str,
    "anki_template": str,
    "anki_tag_filter": str,
    "anki_tag_render_filter": str,
    "json_dict_template": str,
    "json_dict_exclude_keys": List,
    "audio_backend": Literal["whisper", "deepgram"],
    "audio_unsilence": bool,
    "whisper_lang": str,
    "whisper_prompt": str,
    "deepgram_kwargs": dict,
    "youtube_language": str,
    "youtube_translation": str,
    "youtube_audio_backend": Literal["youtube", "whisper", "deepgram"],
    "load_functions": List,
    "doccheck_min_token": int,
    "doccheck_max_token": int,
    "doccheck_min_lang_prob": float,
    "online_media_url_regex": str,
    "online_media_resourcetype_regex": str,
    "loading_failure": str,
}

# extra arguments supported when instanciating wdoc
extra_args_types = {
    "path": Union[str, Path],
    "embed_instruct": str,
    "out_file": Union[str, Path],
    "include": str,
    "exclude": str,
    "filter_content": Union[List[str], str],
    "filter_metadata": Union[List[str], str],
    "source_tag": str,
    "pattern": str,
    "recursed_filetype": str,
}
extra_args_types.update(filetype_arg_types)


class DocDict(dict):
    """like dictionnaries but only allows keys that can be used when loading
    a document. Also checks the value type.

    The environnment variable 'WDOC_STRICT_DOCDICT' is a default value
    at instanciation time.
    Depending on WDOC_STRICT_DOCDICT (if not passed manually):
        if True: crash if unexpected arg
        if False: print in red if unexpected arg but add anyway
        if "strip": print in red but don't add
    """

    allowed_keys: set = set(
        sorted(
            [
                "path",
                "filetype",
                "file_hash",
                "source_tag",
                "recur_parent_id",
            ]
            + list(filetype_arg_types.keys())
        )
    )
    allowed_types: dict = filetype_arg_types
    __strict__ = WDOC_STRICT_DOCDICT

    def __hash__(self):
        "make it hashable, to check for duplicates"
        keys = sorted(self.keys())

        as_string = ""
        for k in keys:
            as_string += "\n"
            as_string += str(k)
            try:
                as_string += jhash(self[k])
            except Exception:
                as_string += str(self[k])
        return hash(as_string)

    def __check_values__(self, key, value, strict) -> bool:
        if key not in self.allowed_keys:
            mess = (
                f"Cannot set key '{key}' in a DocDict. Allowed keys are "
                f"'{','.join(self.allowed_keys)}'\nYou can use the env "
                "variable WDOC_STRICT_DOCDICT to avoid this issue."
            )
            if strict is True:
                raise UnexpectedDocDictArgument(mess)
            elif strict is False:
                red(mess)
                return True
            elif strict == "strip":
                red(mess)
                return False
            else:
                raise ValueError(strict)

        elif (
            (key in self.allowed_types)
            and (value is not None)
            and (not is_bearable(value, self.allowed_types[key]))
        ):
            mess = (
                f"Type of key {key} should be {self.allowed_types[key]},"
                f"not {type(value)}."
                "\nYou can use the env "
                "variable WDOC_STRICT_DOCDICT to avoid this issue."
            )

            if strict is True:
                raise UnexpectedDocDictArgument(mess)
            elif strict is False:
                red(mess)
                return True
            elif strict == "strip":
                red(mess)
                return False
            else:
                raise ValueError(strict)
        return True

    def __init__(self, docdict: dict, strict=WDOC_STRICT_DOCDICT) -> None:
        assert docdict, "Can't give an empty docdict as argument"
        assert strict in [True, False, "strip"], "Unexpected strict value"
        ignore_kwargs = []
        for k, v in docdict.items():
            if not self.__check_values__(k, v, strict):
                ignore_kwargs.append(k)

        for ik in ignore_kwargs:
            if ik in docdict:
                del docdict[ik]

        if strict != "strip":
            assert docdict, "Can't create DocDict: no args nor kwargs after filtering!"
        super().__init__(docdict)

        self.__strict__ = strict

    def __setitem__(self, key, value) -> None:
        assert self.__strict__ in [True, False, "strip"], "Unexpected strict value"
        self.__check_values__(key, value, self.__strict__)
        super().__setitem__(key, value)


@optional_typecheck
def optional_strip_unexp_args(func: Callable) -> Callable:
    """if the environment variable WDOC_STRICT_DOCDICT is set to 'true'
    then this automatically removes any unexpected argument before calling a
    loader function for a specific filetype."""
    if not WDOC_STRICT_DOCDICT:
        return optional_typecheck(func)
    else:
        # find the true function, otherwise func can be a decorated truefunc and might forget the annotations.
        if hasattr(func, "func"):
            truefunc = func.func
        else:
            truefunc = func
        while hasattr(truefunc, "func"):
            truefunc = truefunc.func

        @optional_typecheck
        @wraps(truefunc)
        def wrapper(*args, **kwargs):
            assert (
                not args
            ), f"We are not expecting args here, only kwargs. Received {args}"
            sig = inspect.signature(truefunc)
            bound_args = sig.bind_partial(**kwargs)

            # Remove unexpected positional arguments
            bound_args.arguments = {
                k: v for k, v in bound_args.arguments.items() if k in sig.parameters
            }

            # Remove unexpected keyword arguments
            kwargs2 = {k: v for k, v in kwargs.items() if k in sig.parameters}

            diffkwargs = {k: v for k, v in kwargs.items() if k not in kwargs2}
            if diffkwargs:
                mess = f"Unexpected args or kwargs in func {func}:"
                for kwarg in diffkwargs:
                    mess += f"\n-KWARG: {kwarg}"
                red(mess)
            assert (
                kwargs2
            ), f"No kwargs2 found for func {func}. There's probably an issue with the decorator"

            return func(**kwargs2)

        return wrapper


@optional_typecheck
def hasher(text: str) -> str:
    """used to hash the text contant of each doc to cache the splitting and
    embeddings"""
    return hashlib.sha256(text.encode()).hexdigest()[:20]


@optional_typecheck
def file_hasher(doc: dict) -> str:
    """used to hash a file's content, as describe by a dict
    A caching mechanism is used to avoid recomputing hash of file that
    have the same path and metadata.
    If the doc dict does not contain a path, the hash of the dict will be
    returned.
    """
    if "path" not in doc:
        return hasher(json.dumps(doc, ensure_ascii=False))
    hashable = False
    if "path" in doc and doc["path"] and Path(doc["path"]).exists():
        hashable = True
    if isinstance(doc["path"], str):
        if doc["path"] == "" or (not doc["path"].strip()):
            hashable = False
    if not doc["path"]:
        hashable = False
    if not isinstance(doc["path"], (str, Path)):
        hashable = False

    if hashable:
        file = Path(doc["path"])
        stats = file.stat()
        return _file_hasher(
            abs_path=str(file.resolve().absolute()),
            stats=[stats.st_mtime, stats.st_ctime, stats.st_ino, stats.st_size],
        )
    else:
        return hasher(json.dumps(doc, ensure_ascii=False))


@optional_typecheck
@hashdoc_cache.cache
def _file_hasher(abs_path: str, stats: List[Union[int, float]]) -> str:
    with open(abs_path, "rb") as f:
        return hashlib.sha256(f.read()).hexdigest()[:20]


@optional_typecheck
def html_to_text(html: str, remove_image: bool = False) -> str:
    """used to strip any html present in the text files"""
    html = html.replace("</li><li>", "<br>")  # otherwise they might get joined
    html = html.replace("</ul><ul>", "<br>")  # otherwise they might get joined
    html = html.replace("<br>", "\n").replace(
        "</br>", "\n"
    )  # otherwise newlines are lost

    soup = bs4.BeautifulSoup(html, "html.parser")
    content = []
    for element in soup.descendants:
        if element.name == "img" and (not remove_image):
            element = str(element)
            if element in html:
                content.append(element)
            elif element[:-2] + ">" in html:
                content.append(element[:-2] + ">")
            else:
                if is_verbose:
                    temptext = " ".join(filter(None, content))
                    red(f"Image not properly parsed from bs4:\n{element}\n{temptext}")
        elif isinstance(element, bs4.NavigableString):
            content.append(str(element).strip())
    text = " ".join(filter(None, content))
    while "\n\n" in text:
        text = text.replace("\n\n", "\n")
    if "<img" in text and remove_image:
        red(f"Failed to remove <img from anki card: {text}")
    return text


@chain
@optional_typecheck
def debug_chain(inputs: Union[dict, List]) -> Union[dict, List]:
    "use it between | pipes | in a chain to open the debugger"
    if hasattr(inputs, "keys"):
        red(str(inputs.keys()))
    breakpoint()
    return inputs


@optional_typecheck
def wrapped_model_name_matcher(model: str) -> str:
    "find the best match for a modelname (wrapped to make some check)"
    # find the currently set api keys to avoid matching models from
    # unset providers
    all_backends = list(litellm.models_by_provider.keys())
    backends = []
    for k, v in dict(os.environ).items():
        if k.endswith("_API_KEY"):
            backend = k.split("_API_KEY")[0].lower()
            if (
                backend not in all_backends
                and is_verbose
                and not printed_unexpected_api_keys[0]
            ):
                yel(
                    f"Found API_KEY for backend {backend} that is not a known backend for litellm."
                )
            else:
                backends.append(backend)
    if is_verbose:
        printed_unexpected_api_keys[0] = True
    assert backends, "No API keys found in environnment"

    # filter by providers
    backend, modelname = model.split("/", 1)
    if backend not in all_backends:
        raise Exception(
            f"Model {model} with backend {backend}: backend not found in "
            "litellm.\nList of litellm providers/backend:\n"
            f"{all_backends}"
        )
    if backend not in backends:
        raise Exception(
            f"Trying to use backend {backend} but no API KEY was found for it in the environnment."
        )
    candidates = litellm.models_by_provider[backend]
    if modelname in candidates:
        return model
    subcandidates = [m for m in candidates if m.startswith(modelname)]
    if len(subcandidates) == 1:
        good = f"{backend}/{subcandidates[0]}"
        return good
    match = get_close_matches(modelname, candidates, n=1)
    if match:
        return match[0]
    else:
        red(
            f"Couldn't match the modelname {model} to any known model. "
            "Continuing but this will probably crash wdoc further "
            "down the code."
        )
        return model


@optional_typecheck
def model_name_matcher(model: str) -> str:
    """find the best match for a modelname (wrapper that checks if the matched
    model has a known cost and print the matched name)
    Bypassed if env variable WDOC_NO_MODELNAME_MATCHING is 'true'
    """
    assert "testing" not in model
    assert "/" in model, f"expected / in model '{model}'"
    if WDOC_NO_MODELNAME_MATCHING:
        whi(f"Bypassing model name matching for model '{model}'")
        return model

    out = wrapped_model_name_matcher(model)
    if out != model and is_verbose:
        yel(f"Matched model name {model} to {out}")
    assert (
        out in litellm.model_cost or out.split("/", 1)[1] in litellm.model_cost
    ), f"Neither {out} nor {out.split('/', 1)[1]} found in litellm.model_cost"
    return out


def get_model_price(model: str) -> List[float]:
    assert (
        not WDOC_ALLOW_NO_PRICE
    ), f"Unexpected value for WDOC_ALLOW_NO_PRICE: {WDOC_ALLOW_NO_PRICE}"
    if model in litellm.model_cost:
        return [
            litellm.model_cost[model]["input_cost_per_token"],
            litellm.model_cost[model]["output_cost_per_token"],
        ]
    elif (trial := model.split("/", 1)[1]) in litellm.model_cost:
        return [
            litellm.model_cost[trial]["input_cost_per_token"],
            litellm.model_cost[trial]["output_cost_per_token"],
        ]
    elif (trial2 := model.split("/")[-1]) in litellm.model_cost:
        return [
            litellm.model_cost[trial2]["input_cost_per_token"],
            litellm.model_cost[trial2]["output_cost_per_token"],
        ]
    else:
        raise Exception(
            red(
                f"Can't find the price of '{model}' nor '{trial}' or '{trial2}'\nUpdate litellm or set WDOC_ALLOW_NO_PRICE=True if you still want to use this model."
            )
        )


@dataclass
class ModelName:
    "Simply stores the different way to phrase a model name"

    original: str
    backend: str = field(init=False)
    model: str = field(init=False)
    sanitized: str = field(init=False)

    def __post_init__(self):
        assert (
            "/" in self.original
        ), f"Modelname must contain a / to distinguish the backend from the model. Received '{self.original}'"
        self.backend, self.model = self.original.split("/", 1)
        self.backend = self.backend.lower()

        # Use a sanitized name for the cache path
        self.sanitized = self.original
        if "/" in self.model:
            try:
                if Path(self.model).exists():
                    with open(
                        Path(self.model).resolve().absolute().__str__(), "rb"
                    ) as f:
                        h = hashlib.sha256(f.read() + str(instruct)).hexdigest()[:15]
                    self.sanitized = Path(self.model).name + "_" + h
            except Exception:
                pass
        self.sanitized = self.sanitized.replace("/", "_")
        if WDOC_PRIVATE_MODE:
            self.sanitized = "private_" + self.sanitized


@optional_typecheck
def get_model_max_tokens(modelname: ModelName) -> int:
    if modelname.original in litellm.model_cost:
        return litellm.model_cost[modelname.original]["max_tokens"]
    elif (trial := modelname.model) in litellm.model_cost:
        return litellm.model_cost[trial]["max_tokens"]
    elif (trial2 := modelname.model.split("/")[-1]) in litellm.model_cost:
        return litellm.model_cost[trial2]["max_tokens"]
    else:
        try:
            return litellm.get_model_info(modelname.original)["max_tokens"]
        except Exception:
            return litellm.get_model_info(modelname.name)[
                "max_tokens"
            ]  # crash if still not found


@optional_typecheck
def get_tkn_length(
    tosplit: str,
    modelname: Union[str, ModelName] = "gpt-3.5-turbo",
) -> int:
    if isinstance(modelname, ModelName):
        modelname = modelname.original
    modelname = modelname.replace("openrouter/", "")
    return litellm.token_counter(model=modelname, text=tosplit)


text_splitters = {}

DEFAULT_SPLITTER_MODELNAME = ModelName("openai/gpt-3.5-turbo")


@optional_typecheck
def get_splitter(
    task: str,
    modelname: ModelName = DEFAULT_SPLITTER_MODELNAME,
) -> TextSplitter:
    "we don't use the same text splitter depending on the task"
    # avoid creating many times this object
    if task not in text_splitters:
        text_splitters[task] = {}
    if modelname.original in text_splitters[task]:
        return text_splitters[task][modelname.original]

    if modelname.original == "testing/testing":
        return get_splitter(task=task, modelname=DEFAULT_SPLITTER_MODELNAME)

    try:
        if modelname.model == "gpt-3.5-turbo":
            max_tokens = 4096
        else:
            max_tokens = get_model_max_tokens(modelname)

        # don't use overly large chunks anyway
        max_tokens = min(max_tokens, WDOC_MAX_CHUNK_SIZE)
    except Exception as err:
        max_tokens = 4096
        red(f"Failed to get max_tokens limit for model {modelname.original}: '{err}'")

    model_tkn_length = partial(get_tkn_length, modelname=modelname.original)

    if task in ["query", "search"]:
        text_splitter = RecursiveCharacterTextSplitter(
            separators=recur_separator,
            chunk_size=int(3 / 4 * max_tokens),  # default 4000
            chunk_overlap=500,  # default 200
            length_function=model_tkn_length,
        )
    elif task in ["summarize_then_query", "summarize"]:
        text_splitter = RecursiveCharacterTextSplitter(
            separators=recur_separator,
            chunk_size=int(1 / 2 * max_tokens),
            chunk_overlap=500,
            length_function=model_tkn_length,
        )
    elif task == "recursive_summary":
        text_splitter = RecursiveCharacterTextSplitter(
            separators=recur_separator,
            chunk_size=int(1 / 4 * max_tokens),
            chunk_overlap=300,
            length_function=model_tkn_length,
        )
    else:
        raise Exception(task)

    text_splitters[task][modelname.original] = text_splitter
    return text_splitter


@optional_typecheck
def check_docs_tkn_length(
    docs: List[Document],
    identifier: str,
    min_token: int = min_token,
    max_token: int = max_token,
    min_lang_prob: float = min_lang_prob,
    check_language: bool = False,
) -> float:
    """checks that the number of tokens in the document is high enough,
    not too low, and has a high enough language probability,
    otherwise something probably went wrong."""
    size = sum([get_tkn_length(d.page_content) for d in docs])
    nline = len("\n".join([d.page_content for d in docs]).splitlines())
    if size <= min_token:
        red(
            f"Example of page from document with too few tokens : {docs[len(docs)//2].page_content}"
        )
        raise Exception(
            f"The number of token from '{identifier}' is {size} <= {min_token}, probably something went wrong?"
        )
    if size >= max_token:
        red(
            f"Example of page from document with too many tokens : {docs[len(docs)//2].page_content}"
        )
        raise Exception(
            f"The number of token from '{identifier}' is {size} >= {max_token}, probably something went wrong?"
        )
    if check_language is False:
        return 1.0

    # check if language check is above a threshold and cast as lowercase as it's apparently what it was trained on
    try:
        probs = [language_detector(d.page_content.replace("\n", "<br>")) for d in docs]
        if probs[0] is None or not probs:
            # bypass if language_detector not defined
            return 1.0
        prob = sum(probs) / len(probs)
        if prob <= min_lang_prob:
            raise Exception(
                f"Low language probability for {identifier}: prob={prob:.3f}<{min_lang_prob}.\nExample page: {docs[len(docs)//2]}"
            )
    except Exception as err:
        if str(err).startswith("Low language probability"):
            raise
        else:
            red(
                f"Error when using language_detector on '{identifier}': {err}. Treating it as valid document."
            )
            return 1.0
    return prob


@optional_typecheck
def unlazyload_modules():
    """make sure no modules are lazy loaded. Useful when we wan't to make
    sure not to loose time and that everything works smoothly. For example
    who knows what happens when multiprocessing with lazy loaded modules."""
    if WDOC_IMPORT_TYPE not in ["both", "lazy"]:
        red("Lazyloading is disabled so not unlazyloading modules.")
        return

    while True:
        found_one = False
        for k, v in sys.modules.items():
            try:
                str(v)
            except Exception as e:
                red(
                    f"Very weird error when loading a package, consider setting WDOC_IMPORT_TYPE to another value than '{WDOC_IMPORT_TYPE}'. Error message was '{e}'"
                )
            if "Lazily-loaded" in str(v):
                try:
                    dir(v)  # this is enough to trigger the loading
                    found_one = True
                except Exception as err:
                    raise Exception(
                        f"Error when unlazyloading module '{k}'. Error: '{err}'"
                        "\nThis can be caused by beartype's typechecking"
                        "\nYou can also try setting the env variable "
                        "WDOC_IMPORT_TYPE to 'native' or 'thread'"
                    ) from err
                break  # otherwise dict size change during iteration
            assert "Lazily-loaded" not in str(v)
        if found_one:
            continue
        else:
            break


@optional_typecheck
def disable_internet(allowed: dict) -> None:
    """
    To be extra sure that no connection goes out of the computer when
    --private is used, we overload the socket module to make it only able to
    reach local connection.
    """
    red(
        "Disabling outgoing internet because private mode is on. "
        "The only allowed IPs from now on are the ones from the "
        "argument llm_api_bases. Note that this permanently filters "
        "outgoing python connections so might interfere with other "
        "python programs is you are importing wdoc instead "
        "of calling it from the shell"
    )

    # unlazyload all modules as otherwise the overloading can happen too late
    unlazyload_modules()

    # list of certainly allowed IPs
    allowed_IPs = set(
        [
            "localhost",
            "127.0.0.1",
        ]
    )
    vals = [
        v.split("//")[1].split(":")[0] if "//" in v else v.split(":")[0]
        for v in list(allowed.values())
    ]
    [allowed_IPs.add(v) for v in vals]

    # list of probably allowed IPs
    private_ranges = [
        ("10.0.0.0", "10.255.255.255"),
        ("172.16.0", "172.31.255.255"),
        ("192.168.0.0", "192.168.255.255"),
        ("127.0.0.0", "127.255.255.255"),
    ]

    @memoize
    @optional_typecheck
    def is_private(ip: str) -> bool:
        "detect if the connection would go to our computer or to a remote server"
        if ip in allowed_IPs:
            return True
        ip = int.from_bytes(socket.inet_aton(ip), "big")
        if ip in allowed_IPs:
            return True
        for start, end in private_ranges:
            if (
                int.from_bytes(socket.inet_aton(start), "big")
                <= ip
                <= int.from_bytes(socket.inet_aton(end), "big")
            ):
                return True
        return False

    @optional_typecheck
    def create_connection(address, *args, **kwargs):
        "overload socket.create_connection to forbid outgoing connections"
        ip = socket.gethostbyname(address[0])
        if not is_private(ip):
            raise RuntimeError("Network connections to the open internet are blocked")
        return socket._original_create_connection(address, *args, **kwargs)

    socket.socket = lambda *args, **kwargs: None
    socket._original_create_connection = socket.create_connection
    socket.create_connection = create_connection

    # sanity check
    assert is_private("localhost")
    assert is_private("10.0.1.32")
    assert is_private("192.168.2.35")
    assert is_private("127.12.13.15")

    # checking allowed ips are okay
    for v in vals:
        assert is_private(v), f"An address failed to be set as private: '{v}'"
    for al in list(allowed.values()):
        ip = socket.gethostbyname(al)
        assert is_private(ip), f"An address failed to be set as private: '{al}'"
    try:
        ip = socket.gethostbyname("www.google.com")
        skip = False
    except Exception as err:
        red(
            "Failed to get IP address of www.google.com to check if it is "
            "indeed blocked. You probably did this on purpose so not "
            f"crashing. Error: '{err}'"
        )
        skip = True
    if not skip:
        assert not is_private(
            ip
        ), f"Failed to set www.google.com as unreachable: IP is '{ip}'"


@optional_typecheck
def set_func_signature(func: Callable) -> Callable:
    """dynamically set the extra args of wdoc.__init__ so that
    instead of **cli_kwargs the signature indicates all allowed arguments.
    Needed to get correct behavior from fire.Fire '--completion'"""
    original_sig = inspect.signature(func)
    assert (
        list(original_sig.parameters.values())[-1].kind == inspect.Parameter.VAR_KEYWORD
    )
    new_params = list(original_sig.parameters.values())[:-1]  # Remove **cli_kwargs
    new_params.extend(
        [
            inspect.Parameter(
                name=arg,
                kind=inspect.Parameter.KEYWORD_ONLY,
                annotation=hint,
                default=None,
            )
            for arg, hint in extra_args_types.items()
        ]
    )
    new_sig = original_sig.replace(parameters=new_params)

    @wraps(func)
    def new_func(self, *args, **kwargs):
        return func(self, *args, **kwargs)

    new_func.__signature__ = new_sig
    new_func.__annotations__ = get_type_hints(func) | extra_args_types

    return new_func


THIN = "<thinking>"
THINE = "</thinking>"
ANSW = "<answer>"
ANSWE = "</answer>"


@optional_typecheck
def thinking_answer_parser(output: str, strict: bool = False) -> dict:
    """separate the <thinking> and <answer> tags in an answer"""
    try:
        # fix </answer> instead of <answer>
        if ANSW not in output and output.count(ANSWE) == 2:
            output = output.replace(ANSWE, ANSW, 1)
        if THIN not in output and output.count(THINE) == 2:
            output = output.replace(THINE, THIN, 1)

        if (THIN not in output) and (ANSW not in output):
            assert (
                THINE not in output
            ), f"Output contains unexpected {THINE}:\n'''\n{output}\n'''"
            assert (
                ANSWE not in output
            ), f"Output contains unexpected {ANSWE}:\n'''\n{output}\n'''"

            return {"thinking": "", "answer": output}

        thinking = ""
        if THIN in output and THINE in output:
            thinking = output.split(THIN, 1)[1].split(THINE, 1)[0].strip()

        answer = ""
        if ANSW in output and ANSWE in output:
            answer = (
                output.replace(thinking, "")
                .split(ANSW, 1)[1]
                .split(ANSWE, 1)[0]
                .strip()
            )

        assert (
            THIN not in answer
        ), f"Parsed answer contained unexpected {THIN}:\n'''\n{output}\n'''"
        assert (
            THINE not in answer
        ), f"Parsed answer contained unexpected {THIN}:\n'''\n{output}\n'''"
        assert (
            ANSW not in answer
        ), f"Parsed answer contained unexpected {ANSW}:\n'''\n{output}\n'''"
        assert (
            ANSWE not in answer
        ), f"Parsed answer contained unexpected {ANSW}:\n'''\n{output}\n'''"

        assert answer, f"No answer could be parsed from LLM output: '{output}'"

        return {"thinking": thinking, "answer": answer}
    except Exception as err:
        if (
            strict
        ):  # otherwise combining answers could snowball into losing lots of text
            raise
        red(
            f"Error when parsing LLM output to get thinking and answer part.\nError: '{err}'\nOriginal output: '{output}'\nWill continue if not using --debug"
        )
        if is_debug:
            raise
        else:
            assert output.strip(), "LLM output was empty"
            return {
                "thinking": "",
                "answer": f"""
<note>
The following LLM answer might have had a problem during parsing
</note>
<output>
{output}
</output>
""".strip(),
            }


# this will contain wdoc's version to be used by langfuse's callback without circular imports
langfuse_callback_holder = []


@optional_typecheck
def create_langfuse_callback(version: str) -> None:
    assert not is_private
    # replace langfuse's env variable if set for wdoc
    for k in [
        "LANGFUSE_PUBLIC_KEY",
        "LANGFUSE_SECRET_KEY",
        "LANGFUSE_HOST",
    ]:
        newk = "WDOC_" + k
        if newk in os.environ and os.environ[newk]:
            os.environ[k] = os.environ[newk]
    if (
        "LANGFUSE_PUBLIC_KEY" in os.environ
        and "LANGFUSE_SECRET_KEY" in os.environ
        and "LANGFUSE_HOST" in os.environ
    ):
        red("Activating langfuse callbacks")
        try:
            # # use litellm's callbacks for chatlitellm backend
            import langfuse
            import litellm

            litellm.success_callback.append("langfuse")
            litellm.failure_callback.append("langfuse")

            # # and use langchain's callback for openai's backend
            # BUT as of october 2024 it seems buggy with chatlitellm, the modelname does not seem to be passed?
            from langfuse.callback import CallbackHandler as LangfuseCallback

            langfuse_callback = LangfuseCallback(
                secret_key=os.environ["LANGFUSE_SECRET_KEY"],
                public_key=os.environ["LANGFUSE_PUBLIC_KEY"],
                host=os.environ["LANGFUSE_HOST"],
                session_id=str(uuid.uuid4()),
                version=version,
            )
            langfuse_callback_holder.append(langfuse_callback)
        except Exception as e:
            red(
                f"Failed to setup langfuse callback, make sure package 'langfuse' is installed. The error was: '{e}'"
            )


@optional_typecheck
def seconds_to_timecode(inp: Union[str, float]) -> str:
    "used for vtt subtitle conversion"
    second = float(inp)
    minute = second // 60
    second = second % 60
    hour = minute // 60
    minute = minute % 60
    hour, minute, second = int(hour), int(minute), int(second)
    return f"{hour:02d}:{minute:02d}:{second:02d}"


@optional_typecheck
def timecode_to_second(inp: str) -> Union[int, float]:
    "turns a vtt timecode into seconds"
    hour, minute, second = map(int, inp.split(":"))
    return hour * 3600 + minute * 60 + second


@optional_typecheck
def is_timecode(inp: Union[float, str]) -> bool:
    try:
        timecode_to_second(inp)
        return True
    except Exception:
        return False


@optional_typecheck
def get_supported_model_params(modelname: ModelName) -> list:
    if modelname.backend == "testing":
        return []
    for test in [
        modelname.original,
        modelname.model,
        model_name_matcher(modelname.original),
    ]:
        params = litellm.get_supported_openai_params(test)
        if params:
            return params
    for test in [
        modelname.original,
        modelname.model,
        model_name_matcher(modelname.original),
    ]:
        params = litellm.get_supported_openai_params(
            test, custom_llm_provider=modelname.backend
        )
        if params:
            return params
    for test in [
        modelname.original,
        modelname.model,
        model_name_matcher(modelname.original),
    ]:
        params = litellm.get_supported_openai_params(
            test, custom_llm_provider="openrouter"
        )
        if params:
            return params
    return []



"""
Sets the default value for environment variables, parse the actual values,
check their types and finally make them easier to access by other parts of
wdoc.
"""

import os
import sys
from functools import wraps

from beartype import BeartypeConf, beartype
from beartype.door import is_bearable
from beartype.typing import Literal, Optional, Union

# must create it because we can't import it from typechecker.py
warn_typecheck = beartype(conf=BeartypeConf(violation_type=UserWarning))


class EnvVar:
    """
    **EXPERIMENTAL**
    This class stores individual environment variables in a way that therre
    values are refreshes if the user modified them. A use case example is to
    have up to date environment values when having multiple users use the
    open-webui wdocTool.
    WDOC_PRIVATE_MODE is excluded from this mechanism, meaning that if it ever is not "False" we never update it for safety reason.
    This should work for ints, floats, str, bool and None but will not work for "isinstance" check.
    """

    def __init__(self, env_name, default_value=None, type_converter=None):
        self.env_name = env_name
        self.default_value = default_value
        self._value = None
        self._type_converter = type_converter

        # Initialize with current env value
        self.__refresh__()

    def __refresh__(self):
        """Get the current value from environment"""
        if self.env_name == "WDOC_PRIVATE_MODE" and self._value != False:
            # if WDOC_PRIVATE_MODE is ever activated, we don't allow deactivating
            return self._value

        env_value = parse(str(os.environ.get(self.env_name)))
        if env_value is None:
            self._value = self.default_value
            return self._value

        if is_bearable(env_value, self._type_converter):
            self._value = env_value
            return self._value

        failed = True
        if isinstance(self._type_converter, type(str)):
            try:
                self._value = self._type_converter(env_value)
            except (ValueError, TypeError):
                pass
        elif isinstance(
            self._type_converter, (type(Optional[str]), type(Literal["a"]))
        ):
            failed = True
            to_try = [self._type_converter]
            if hasattr(self._type_converter, "__args__"):
                to_try.extend(self._type_converter.__args__)
            for t in to_try:
                if t is type(None):
                    continue
                if not callable(t):
                    t = type(t)
                try:
                    assert callable(
                        t
                    ), f"Type '{t}' is not callable but we expect it to be"
                    self._value = t(env_value)
                    failed = False
                    break
                except (ValueError, TypeError):
                    pass
        else:
            raise Exception(
                f"Couldn't handle the type '{self._type_converter}' for env variable '{self.env_name}' with new value '{env_value}'"
            )

        if failed:
            raise TypeError(
                f"Couldn't set env variable value '{env_value}' of '{self.env_name}' to type '{self._type_converter}'"
            )
        else:
            self._value = env_value
            return self._value

    def __getattr__(self, name):
        # Check if env has changed
        current_value = self.__refresh__()
        if name == "__refresh__":
            return current_value
        else:
            print(name)
            return getattr(self._value, name)(*args, **kwargs)

    def __eq__(self, other):
        self.__refresh__()
        return self._value.__eq__(other)

    def __ne__(self, other):
        self.__refresh__()
        return self._value.__ne__(other)

    def __lt__(self, other):
        self.__refresh__()
        return self._value.__lt__(other)

    def __le__(self, other):
        self.__refresh__()
        return self._value.__le__(other)

    def __gt__(self, other):
        self.__refresh__()
        return self._value.__gt__(other)

    def __ge__(self, other):
        self.__refresh__()
        return self._value.__ge__(other)

    def __str__(self):
        self.__refresh__()
        return self._value.__str__()

    def __repr__(self):
        self.__refresh__()
        return self._value.__repr__()

    def __int__(self):
        self.__refresh__()
        return self._value.__int__()

    def __float__(self, other):
        self.__refresh__()
        return self._value.__float__(other)

    def __add__(self, other):
        self.__refresh__()
        return self._value.__add__(other)

    def __isinstance__(self, cls):
        return isinstance(self._value, cls)

    def __instancecheck__(self, instance):
        return isinstance(self._value, instance)


@warn_typecheck
def parse(val: str) -> Optional[Union[bool, int, str]]:
    if val.lower() == "true":
        return True
    elif val.lower() == "false":
        return False
    elif val.isdigit():
        return int(val)
    elif val.lower() == "none" or val == "":
        return None
    else:
        return val


WDOC_TYPECHECKING = "warn"
WDOC_NO_MODELNAME_MATCHING = True
WDOC_ALLOW_NO_PRICE = False
WDOC_OPEN_ANKI = False
WDOC_STRICT_DOCDICT = False
WDOC_MAX_LOADER_TIMEOUT = -1
WDOC_MAX_PDF_LOADER_TIMEOUT = -1  # disabled as it can make the parsing slower
WDOC_PRIVATE_MODE = False
WDOC_DEBUGGER = False
WDOC_EXPIRE_CACHE_DAYS = 0
WDOC_EMPTY_LOADER = False
WDOC_BEHAVIOR_EXCL_INCL_USELESS = "warn"
WDOC_IMPORT_TYPE = "thread"
WDOC_MOD_FAISS_SCORE_FN = False
WDOC_LLM_MAX_CONCURRENCY = 15
WDOC_SEMANTIC_BATCH_MAX_TOKEN_SIZE = 1000
WDOC_MAX_CHUNK_SIZE = 16_000

WDOC_DEFAULT_MODEL = "anthropic/claude-3-7-sonnet-20250219"
WDOC_DEFAULT_EMBED_MODEL = "openai/text-embedding-3-small"
WDOC_DEFAULT_EMBED_DIMENSION = None
WDOC_EMBED_TESTING = True
WDOC_DEFAULT_QUERY_EVAL_MODEL = "anthropic/claude-3-5-haiku-20241022"

WDOC_LANGFUSE_PUBLIC_KEY = None
WDOC_LANGFUSE_SECRET_KEY = None
WDOC_LANGFUSE_HOST = None

WDOC_LITELLM_TAGS = None
WDOC_LITELLM_USER = "wdoc_llm"
WDOC_ENABLE_EXPERIMENTAL_ENV = False

# by default use lazy loading if using --help argument
if " --help" in " ".join(sys.argv):
    print("--help so using lazy loading")
    WDOC_IMPORT_TYPE = "lazy"

valid_types = {
    "WDOC_TYPECHECKING": Literal["disabled", "warn", "crash"],
    "WDOC_NO_MODELNAME_MATCHING": bool,
    "WDOC_ALLOW_NO_PRICE": bool,
    "WDOC_OPEN_ANKI": bool,
    "WDOC_STRICT_DOCDICT": Union[bool, Literal["strip"]],
    "WDOC_MAX_LOADER_TIMEOUT": int,
    "WDOC_MAX_PDF_LOADER_TIMEOUT": int,
    "WDOC_PRIVATE_MODE": bool,
    "WDOC_DEBUGGER": bool,
    "WDOC_EXPIRE_CACHE_DAYS": int,
    "WDOC_EMPTY_LOADER": bool,
    "WDOC_BEHAVIOR_EXCL_INCL_USELESS": Literal["warn", "crash"],
    "WDOC_IMPORT_TYPE": Literal["native", "lazy", "thread", "both"],
    "WDOC_MOD_FAISS_SCORE_FN": bool,
    "WDOC_LLM_MAX_CONCURRENCY": int,
    "WDOC_SEMANTIC_BATCH_MAX_TOKEN_SIZE": int,
    "WDOC_MAX_CHUNK_SIZE": int,
    "WDOC_DEFAULT_MODEL": str,
    "WDOC_DEFAULT_EMBED_MODEL": str,
    "WDOC_DEFAULT_EMBED_DIMENSION": Optional[int],
    "WDOC_EMBED_TESTING": bool,
    "WDOC_DEFAULT_QUERY_EVAL_MODEL": str,
    "WDOC_LANGFUSE_PUBLIC_KEY": Optional[str],
    "WDOC_LANGFUSE_SECRET_KEY": Optional[str],
    "WDOC_LANGFUSE_HOST": Optional[str],
    "WDOC_LITELLM_TAGS": Optional[str],
    "WDOC_LITELLM_USER": str,
    "WDOC_ENABLE_EXPERIMENTAL_ENV": bool,
}

# sanity check for the default values
for k, v in locals().copy().items():
    if not k.startswith("WDOC_"):
        continue
    assert k in valid_types, k
    assert is_bearable(v, valid_types[k]), v

# store the env variable instead of the default values but check their types
for k in os.environ.keys():
    if not k.lower().startswith("wdoc_"):
        continue
    v = parse(os.environ[k])
    # assert k in locals().keys(), f"Unexpected key env variable starting by 'wdoc_': {k}."
    if k not in locals().keys():
        print(
            f"Unexpected key env variable starting by 'wdoc_': {k}. This might be a typo in your configuration!"
        )
    else:
        assert is_bearable(
            v, valid_types[k]
        ), f"Unexpected type of env variable '{k}': '{type(v)}' but expected '{valid_types['k']}'"
        locals()[k] = v


# set them as EnvVar
if WDOC_ENABLE_EXPERIMENTAL_ENV:
    for k, v in locals().copy().items():
        if not k.startswith("WDOC_"):
            continue
        assert k in valid_types, k
        assert is_bearable(v, valid_types[k]), v
        locals()[k] = EnvVar(k, v, valid_types[k])



"""
easy access for all other files wether we are in verbose mode or not etc
"""

import os
import platform
import sys

is_linux = platform.system() == "Linux"

cmdline = " ".join(sys.argv[1:])


def check_kwargs(arg: str, abbrv: str = None) -> bool:
    if f" {arg}" in cmdline or f" --{arg}" in cmdline:
        return True
    if abbrv and f" -{abbrv}" in cmdline:
        return True
    return False


is_debug = check_kwargs("debug", "d")

is_verbose = is_debug or check_kwargs("verbose", "v")

is_silent = check_kwargs("silent", "s")

md_printing_disabled = check_kwargs("disable_md_printing")


class PrivateSanityChecker(int):
    "simple class that ALWAYS checks that WDOC_PRIVATE_MODE is appropriate when is_private is compared to anything"

    def __new__(cls, value=False):  # needed to subclass bool
        return super().__new__(cls, bool(value))

    def __init__(self, value):
        assert isinstance(value, bool)
        self.value = value

    def __sanity_check__(self):
        if self.value:
            assert str(os.environ["WDOC_PRIVATE_MODE"]).lower() == "true"
        else:
            assert str(os.environ["WDOC_PRIVATE_MODE"]).lower() == "false"

    def __eq__(self, other):
        self.__sanity_check__()
        return self.value == other


is_private = PrivateSanityChecker(check_kwargs("private"))

# useful to know if we should use tqdm or not (it can cause broken pipe errors
# otherwise) and modify the formatting output
is_piped = sys.stdout.isatty()



"""
Code related to the prompt (in the sense of "directly ask the user a question")
"""

import json
import time
from pathlib import Path
from textwrap import dedent

from beartype.typing import Any, Optional, Tuple
from prompt_toolkit import PromptSession
from prompt_toolkit.completion import Completer, Completion, WordCompleter
from prompt_toolkit.formatted_text import FormattedText

from .logger import md_printer, red, whi
from .misc import cache_dir
from .typechecker import optional_typecheck


@optional_typecheck
def get_toolbar_text(settings: dict) -> Any:
    """Parse settings for display in the prompt toolbar.

    Args:
        settings (dict): Dictionary containing the current settings

    Returns:
        Any: Formatted text suitable for display in the toolbar
    """
    out = []
    keys = sorted(list(settings.keys()))
    for k in keys:
        if k == "task":
            continue
        v = settings[k]
        out.append(f"{k.replace('_', ' - ').title()}: {v}")
    out = ["class:toolbar"] + [" - ".join(out)]
    return FormattedText([tuple(out)])


class SettingsCompleter(Completer):
    @optional_typecheck
    def __init__(
        self, wdocCliSettings, wdocHistoryPrompts, wdocHistoryWords, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.wdocCliSettings = wdocCliSettings
        self.wdocHistoryPrompts = wdocHistoryPrompts
        self.wdocHistoryWords = wdocHistoryWords

    @optional_typecheck
    def get_completions(self, document, complete_event):
        text = document.text_before_cursor
        if not text.strip():
            yield Completion("/debug", start_position=-len(text))
            yield Completion("/settings", start_position=-len(text))
            yield Completion("/help", start_position=-len(text))
        elif text.startswith("/"):
            if "/debug".startswith(text):
                yield Completion("/debug", start_position=-len(text))
            if "/help".startswith(text):
                yield Completion("/help", start_position=-len(text))
            if "/settings ".startswith(text) or "/settings " in text:
                settings = sorted(list(self.wdocCliSettings.keys()))
                for setting in settings:
                    if setting == "task":
                        continue
                    compl = f"/settings {setting}={self.wdocCliSettings[setting]}"
                    if compl.startswith(text):
                        yield Completion(compl, start_position=-len(text))
        else:
            # words autocompletion
            if " " in text and not text.endswith(" "):
                last_word = text.split(" ")[-1]
                word_cnt = 0
                for word in self.wdocHistoryWords:
                    if word_cnt >= 3:
                        break
                    if word.lower().startswith(last_word.lower()):
                        yield Completion(word, start_position=-len(last_word))
                        word_cnt += 1

            # entire prompt autocompletion
            for hist in self.wdocHistoryPrompts:
                if hist.lower().startswith(text.lower()):
                    yield Completion(hist, start_position=-len(text))


@optional_typecheck
def show_help() -> None:
    """Display CLI help information.

    This function displays the CLI help information by formatting and showing
    the docstring from the ask_user function.

    Returns:
        None
    """
    md_printer(dedent(ask_user.__doc__).strip())


@optional_typecheck
def ask_user(settings: dict) -> Tuple[str, dict]:
    """
    ## Command line manual
    * **Available Commands:**
        * /help or ?
        * /debug
        * /settings (syntax: '/settings top_k=5')
    * **Settings keys and values:**
        * top_k: int > 0
        * multiline: boolean
        *retriever: a string containing '_' separated retriever from the
        following list:
            * 'default' to use regular embedding search
            * 'knn' to use KNN
            * 'svm' to use SVM
            * 'multiquery' to use Hypothetical Document Embedding search
            * 'parent' to use parent retriever
        To use several '/settings retriever=knn_svm_default'
        * relevancy: float, from set [-1:+1]
    * **Tips:**
        * Each LLM used has a nickname: use it to adress specific instructions.
          The nicknames are "Summarizer", "Evaluator", "Answerer" and "Combiner".
        * In multiline mode, use ctrl+D to send the text (sometimes
        multiple times).
        * For more information run 'wdoc --help'
        * History is saved and shared across all runs
        * If you use '>>>>' once in the middle of your text, the left part will be
        used as a query find the documents and the right part will be the
        question to answer. For example: 'tuberculosis among medical students
        in the 20th century >>>> what are the statistics about epidemiology
        of tuberculosis among medical students in the 20th century?'. This is
        not always useful but in some cases depending on documents and
        retriever it can be needed to avoid having to set top_k too high.
    """
    md_printer("# wdoc Prompt")

    # loading history from files
    prev_questions = []
    pp_file = cache_dir / "query_history.json"
    if pp_file.exists():
        pp_list = json.load(pp_file.open("r"))
        assert isinstance(pp_list, list), "Invalid cache type"
        for i, pp in enumerate(pp_list):
            assert isinstance(pp, dict), "Invalid item in cache"
            assert "prompt" in pp, "Invalid item in cache"
        for pp in pp_list:
            if "timestamp" not in pp:
                pp["timestamp"] = 0
            if "task" not in pp:
                pp["task"] = "query"
            if pp not in prev_questions:
                prev_questions.append(pp)
        prev_questions = sorted(
            prev_questions,
            key=lambda x: x["timestamp"],
        )

    prompts = [x["prompt"] for x in prev_questions if x["task"] == settings["task"]]
    words = [w for w in " ".join(prompts).split(" ") if len(w) > 2 and w.isalpha()]
    completer = SettingsCompleter(
        wdocCliSettings=settings, wdocHistoryPrompts=prompts, wdocHistoryWords=words
    )

    while True:
        if settings["multiline"]:
            whi("Multiline mode enabled. Use ctrl+D to send.")
        session = PromptSession(
            bottom_toolbar=lambda: get_toolbar_text(settings),
            completer=completer,
        )
        try:
            user_input = session.prompt(
                "> ",
                # completer=autocomplete,
                vi_mode=True,
                multiline=settings["multiline"],
            )
        except KeyboardInterrupt:
            red("Quitting.")
            raise SystemExit()
        except EOFError:
            if settings["multiline"]:
                pass
            else:
                red("Quitting.")
                raise SystemExit()
        user_input = user_input.strip()

        # quit
        if user_input.strip() in ["quit", "Q", "q"]:
            whi("Quitting.")
            raise SystemExit()
        elif user_input == "/debug":
            whi("Entering debug mode.")
            breakpoint()
            whi("Going back to the prompt.")
            continue
        elif user_input in ["/help", "?"]:
            show_help()
            continue

        # handle settings
        if user_input.startswith("/settings "):
            if "=" not in user_input:
                red("Invalid settings syntax: missing '='")
                show_help()
                continue
            input_sett = user_input.split(" ")
            if not input_sett[0] == "/settings":
                red("Invalid settings syntax: does not start with '/settings '")
                show_help()
                continue
            if not len(input_sett) == 2:
                red("Invalid settings syntax: too many spaces")
                show_help()
                continue
            input_sett = input_sett[1]
            input_sett = input_sett.split("=")
            if not len(input_sett) == 2:
                red("Invalid settings syntax: expected one '=' symbol")
                show_help()
                continue
            sett_k, sett_v = input_sett
            if sett_k not in settings.keys():
                red("Invalid settings: '{sett_k}' is not a valid setting key")
                show_help()
                continue
            if settings[sett_k] == sett_v:
                red("Invalid settings: '{sett_k}' is already has value '{sett_v}'")
                show_help()
                continue
            try:
                if sett_k == "top_k":
                    assert int(sett_v) > 0, f"Can't set top_k to <= 0 ({sett_v})"
                elif sett_k == "relevancy":
                    assert (
                        float(sett_v) >= -1 and float(sett_v) <= 1
                    ), f"Can't set relevancy to < -1 or > +1 ({sett_v})"
                    sett_v = float(sett_v)
                elif sett_k == "retriever":
                    assert all(
                        retriev in ["default", "multiquery", "knn", "svm", "parent"]
                        for retriev in sett_v.split("_")
                    ), f"Invalid retriever value: {sett_v}"
                elif sett_k == "multiline":
                    if sett_v.lower() == "true":
                        sett_v = True
                    elif sett_v.lower() == "false":
                        sett_v = False
                    sett_v = bool(sett_v)
                settings[sett_k] = type(settings[sett_k])(sett_v)
            except Exception as err:
                red(
                    f"Error: can't set '{sett_k}' to '{sett_v}' because it "
                    f"can't keep the type '{type(settings[sett_k])}'\n"
                    f"Error: '{err}'"
                )
                show_help()
                continue
            whi(f"Set {sett_k}={sett_v}")
            continue
        elif "/settings" in user_input:
            red(f"Detected '/settings' but not at the start, retrying.")
            show_help()
            continue

        break
    md_printer("### Done prompting")

    # saving new history to file
    if len([x for x in prev_questions if x["prompt"].strip() == user_input]) == 0:
        prev_questions.append(
            {
                "prompt": user_input,
                "timestamp": int(time.time()),
                "task": settings["task"],
            }
        )
    prev_questions = sorted(
        prev_questions,
        key=lambda x: x["timestamp"],
    )
    temp_file = Path(str(pp_file.resolve().absolute()) + ".temp")
    json.dump(prev_questions, temp_file.open("w"), indent=4)
    assert temp_file.exists()
    pp_file.unlink(missing_ok=True)
    temp_file.rename(pp_file)

    return user_input, settings



"""
Retrievers used to retrieve the appropriate embeddings for a given query.
"""

from beartype.typing import Any, List, Union
from langchain.docstore.document import Document
from langchain.retrievers import ParentDocumentRetriever
from langchain.retrievers.multi_query import MultiQueryRetriever
from langchain.storage import LocalFileStore
from langchain_community.chat_models import ChatLiteLLM
from langchain_core.retrievers import BaseRetriever
from langchain_openai import ChatOpenAI

from .misc import cache_dir, get_splitter
from .prompts import multiquery_parser, prompts
from .typechecker import optional_typecheck


@optional_typecheck
def create_multiquery_retriever(
    llm: Union[ChatLiteLLM, ChatOpenAI],
    retriever: BaseRetriever,
) -> MultiQueryRetriever:
    # advanced mode using pydantic parsers
    llm_chain = prompts.multiquery | llm | multiquery_parser
    mqr = MultiQueryRetriever(
        retriever=retriever,
        llm_chain=llm_chain,
    )

    # TODO: fix the fallback: the llm_chain has to have a callback instead
    # # as pydantic parsing can be complicated for some model
    # # we keep the default multi query retriever as a fallback
    # default = MultiQueryRetriever.from_llm(
    #     retriever=retriever,
    #     llm=llm,
    # )
    # resilient = mqr.with_fallbacks(fallbacks=[default])
    # return resilient
    return mqr


@optional_typecheck
def create_parent_retriever(
    task: str,
    loaded_embeddings: Any,
    loaded_docs: List[Document],
    top_k: int,
    relevancy: float,
) -> BaseRetriever:
    "https://python.langchain.com/docs/modules/data_connection/retrievers/parent_document_retriever"
    csp = get_splitter(task)
    psp = get_splitter(task)
    psp._chunk_size *= 4
    parent = ParentDocumentRetriever(
        vectorstore=loaded_embeddings,
        docstore=LocalFileStore(cache_dir / "parent_retriever"),
        child_splitter=csp,
        parent_splitter=psp,
        search_type="similarity",
        search_kwargs={
            "k": top_k,
            "score_threshold": relevancy,
        },
    )
    parent.add_documents(loaded_docs)
    return parent



import os
import sys
import subprocess
import tempfile
from pathlib import Path
from copy import copy

import pytest
from langchain_core.documents.base import Document

os.environ["WDOC_TYPECHECKING"] = "crash"

# Default model names if not specified in environment
WDOC_TEST_OLLAMA_EMBED_MODEL = os.getenv(
    "WDOC_TEST_OLLAMA_EMBED_MODEL", "snowflake-arctic-embed2"
)
WDOC_TEST_OPENAI_MODEL = os.getenv("WDOC_TEST_OPENAI_MODEL", "gpt-4o")
WDOC_TEST_OPENAI_EVAL_MODEL = os.getenv("WDOC_TEST_OPENAI_EVAL_MODEL", "gpt-4o-mini")
WDOC_TEST_OPENAI_EMBED_MODEL = os.getenv(
    "WDOC_TEST_OPENAI_EMBED_MODEL", "text-embedding-3-small"
)

from wdoc.wdoc import wdoc
from wdoc.utils.env import WDOC_TYPECHECKING, WDOC_ENABLE_EXPERIMENTAL_ENV
from wdoc.utils.misc import ModelName
from wdoc.utils.embeddings import load_embeddings_engine
from wdoc.utils.embeddings import test_embeddings as _test_embeddings
from wdoc.utils.tasks.query import semantic_batching
from wdoc.utils.embeddings import load_embeddings_engine
from wdoc.utils.misc import ModelName
from wdoc.utils.env import WDOC_DEFAULT_EMBED_MODEL


@pytest.mark.basic
def test_wdoc_version():
    """Test that wdoc has a valid version string."""
    assert isinstance(wdoc.VERSION, str)
    assert len(wdoc.VERSION.split(".")) == 3


@pytest.mark.basic
def test_fail_parse_small_file_text(sample_text_file):
    """Test that a too small text file parsing fails."""
    # should fail because the file is too small
    with pytest.raises(Exception):
        wdoc.parse_file(
            path=str(sample_text_file), filetype="txt", debug=False, verbose=False
        )


@pytest.mark.basic
def test_parse_file_text(sample_text_file):
    """Test basic text file parsing."""
    # make a bigger text file
    f = Path(sample_text_file)
    content = f.read_text()
    f.write_text(50 * (content + "\n"))
    docs = wdoc.parse_file(
        path=str(sample_text_file),
        filetype="txt",
        debug=False,
        verbose=False,
        format="langchain",
    )
    assert len(docs) > 0
    assert docs[0].page_content.startswith("This is a test document")
    assert "multiple lines" in docs[0].page_content


@pytest.mark.basic
def test_parse_file_formats(sample_text_file):
    """Test text-only output from parse_file."""
    f = Path(sample_text_file)
    content = f.read_text()
    f.write_text(50 * (content + "\n"))
    docs = wdoc.parse_file(
        path=str(sample_text_file),
        filetype="txt",
        format="langchain",
        debug=False,
        verbose=False,
    )
    assert isinstance(docs, list), type(docs)
    assert len(docs) == 1, len(docs)
    assert all(isinstance(d, Document) for d in docs), ",".join(type(d) for d in docs)
    doc = docs[0]
    assert isinstance(doc, Document), type(doc)
    assert doc.page_content.startswith("This is a test document"), doc
    assert "multiple lines" in doc.page_content, doc.page_content

    ld = wdoc.parse_file(
        path=str(sample_text_file),
        filetype="txt",
        format="langchain_dict",
        debug=False,
        verbose=False,
    )
    assert isinstance(ld, list), type(ld)
    for ldd in ld:
        assert isinstance(ldd, dict), ldd
        assert "page_content" in ldd, ldd
        assert "metadata" in ldd, ldd

    text = wdoc.parse_file(
        path=str(sample_text_file),
        filetype="txt",
        format="text",
        debug=False,
        verbose=False,
    )
    assert isinstance(text, str), type(text)

    xml = wdoc.parse_file(
        path=str(sample_text_file),
        filetype="txt",
        format="xml",
        debug=False,
        verbose=False,
    )
    assert isinstance(xml, str), type(xml)

    assert xml != text


@pytest.mark.basic
def test_invalid_filetype():
    """Test that invalid filetype raises an error."""
    with pytest.raises(Exception):
        wdoc.parse_file(
            path="dummy.txt", filetype="invalid_type", debug=False, verbose=False
        )


@pytest.mark.basic
def test_parse_online_pdf():
    """Test parsing an online PDF about situational awareness."""
    docs = wdoc.parse_file(
        path="https://situational-awareness.ai/wp-content/uploads/2024/06/situationalawareness.pdf",
        filetype="online_pdf",
        format="langchain",
        debug=False,
        verbose=False,
    )
    assert isinstance(docs, list)
    assert len(docs) > 0
    assert all(isinstance(d, Document) for d in docs)
    assert any("alphago" in d.page_content.lower() for d in docs)


@pytest.mark.api
@pytest.mark.skipif(
    " -m api" not in " ".join(sys.argv),
    reason="Skip tests using external APIs by default, use '-m api' to run them.",
)
def test_summary_tim_urban():
    """Test summarization of Tim Urban's procrastination video."""
    _ = wdoc(
        task="summarize",
        path="https://www.youtube.com/watch?v=arj7oStGLkU",
        model=f"openai/{WDOC_TEST_OPENAI_MODEL}",
        query_eval_model=f"openai/{WDOC_TEST_OPENAI_EVAL_MODEL}",
        embed_model=f"openai/{WDOC_TEST_OPENAI_EMBED_MODEL}",
        # filetype="youtube",
        filetype="auto",
        debug=False,
        verbose=False,
        import_mode=True,
    )


@pytest.mark.api
@pytest.mark.skipif(
    " -m api" not in " ".join(sys.argv),
    reason="Skip tests using external APIs by default, use '-m api' to run them.",
)
def test_query_tim_urban():
    """Test query task on Tim Urban's procrastination video."""
    _ = wdoc(
        task="query",
        query="What university did the author go to?",
        path="https://www.youtube.com/watch?v=arj7oStGLkU",
        model=f"openai/{WDOC_TEST_OPENAI_MODEL}",
        query_eval_model=f"openai/{WDOC_TEST_OPENAI_EVAL_MODEL}",
        embed_model=f"openai/{WDOC_TEST_OPENAI_EMBED_MODEL}",
        # filetype="youtube",
        filetype="auto",
        debug=False,
        verbose=False,
        import_mode=True,
    )


@pytest.mark.api
@pytest.mark.skipif(
    " -m api" not in " ".join(sys.argv),
    reason="Skip tests using external APIs by default, use '-m api' to run them.",
)
def test_whisper_tim_urban():
    """Test summarization of Tim Urban's video using whisper transcription."""
    _ = wdoc(
        task="summarize",
        path="https://www.youtube.com/watch?v=arj7oStGLkU",
        model=f"openai/{WDOC_TEST_OPENAI_MODEL}",
        query_eval_model=f"openai/{WDOC_TEST_OPENAI_EVAL_MODEL}",
        embed_model=f"openai/{WDOC_TEST_OPENAI_EMBED_MODEL}",
        filetype="youtube",
        youtube_audio_backend="whisper",
        whisper_lang="en",
        debug=False,
        verbose=False,
        import_mode=True,
    )


@pytest.mark.api
@pytest.mark.skipif(
    " -m api" not in " ".join(sys.argv),
    reason="Skip tests using external APIs by default, use '-m api' to run them.",
)
def test_openai_embeddings():
    emb = load_embeddings_engine(
        modelname=ModelName(f"openai/{WDOC_TEST_OPENAI_EMBED_MODEL}"),
        cli_kwargs={},
        api_base=None,
        embed_kwargs={},
        private=False,
        do_test=True,
    )
    _test_embeddings(emb)


@pytest.mark.basic
def test_ollama_embeddings():
    emb = load_embeddings_engine(
        modelname=ModelName(f"ollama/{WDOC_TEST_OLLAMA_EMBED_MODEL}"),
        cli_kwargs={},
        api_base=None,
        embed_kwargs={},
        private=False,
        do_test=True,
    )
    _test_embeddings(emb)


@pytest.mark.basic
def test_help_output_shell():
    """Test that --help output contains expected docstring."""
    result = subprocess.run(
        ["wdoc", "--help"],
        capture_output=True,
        text=True,
        check=False,
        # stderr=subprocess.STDOUT,
    )
    output = result.stdout + result.stderr
    print(output)
    assert (
        "This docstring is dynamically updated with the content of wdoc/docs/help.md"
        not in output
    )
    assert "Content of wdoc/docs/help.md" in output


@pytest.mark.basic
def test_help_output_python():
    """Test that --help output contains expected docstring."""
    result = subprocess.run(
        ["python", "-m", "wdoc", "--help"], capture_output=True, text=True, check=False
    )
    output = result.stdout + result.stderr
    assert (
        "This docstring is dynamically updated with the content of wdoc/docs/help.md"
        not in output
    )
    assert "Content of wdoc/docs/help.md" in output


@pytest.mark.basic
def test_parse_file_help_output_shell():
    """Test that --help output contains expected docstring."""
    result = subprocess.run(
        ["wdoc", "parse", "--help"], capture_output=True, text=True, check=False
    )
    output = result.stdout + result.stderr
    assert (
        "This docstring is dynamically updated with the content of wdoc/docs/parse_file_help.md"
        not in output
    )
    assert "Content of wdoc/docs/parse_file_help.md" in output


@pytest.mark.basic
def test_parse_file_help_output_python():
    """Test that --help output contains expected docstring."""
    result = subprocess.run(
        ["python", "-m", "wdoc", "parse", "--help"],
        capture_output=True,
        text=True,
        check=False,
    )
    output = result.stdout + result.stderr
    assert (
        "This docstring is dynamically updated with the content of wdoc/docs/parse_file_help.md"
        not in output
    )
    assert "Content of wdoc/docs/parse_file_help.md" in output


@pytest.mark.basic
def test_semantic_batching():
    """Test that semantic_batching properly groups related texts."""
    texts = [
        "The cat chased the mouse around the house",
        "Python is a popular programming language",
        "JavaScript is used for web development",
        "The dog barked at the mailman yesterday",
    ]

    embeddings = load_embeddings_engine(
        modelname=ModelName(WDOC_DEFAULT_EMBED_MODEL),
        cli_kwargs={},
        api_base=None,
        embed_kwargs={},
        private=False,
        do_test=True,
    )

    batches = semantic_batching(texts, embeddings)

    # Basic validation
    assert isinstance(batches, list)
    assert len(batches) >= 1
    assert all(isinstance(batch, list) for batch in batches)

    # Check all texts are present
    all_texts = []
    for batch in batches:
        all_texts.extend(batch)
    assert sorted(all_texts) == sorted(texts)

    # Check semantic grouping (programming languages should be together)
    for batch in batches:
        if "Python" in batch[0]:
            assert any("JavaScript" in text for text in batch)
        elif "JavaScript" in batch[0]:
            assert any("Python" in text for text in batch)

    assert texts != all_texts and texts != all_texts[::-1], all_texts


@pytest.mark.basic
def test_parse_docx():
    """Test parsing a DOCX file."""
    # Create temporary file and download sample DOCX file
    with tempfile.NamedTemporaryFile(suffix=".docx", delete=False) as tmp:
        tmp_path = tmp.name

    url = "https://freetestdata.com/wp-content/uploads/2021/09/Free_Test_Data_500KB_DOCX.docx"
    subprocess.run(["wget", "-q", url, "-O", tmp_path], check=True)

    try:
        # Parse the file
        docs = wdoc.parse_file(
            path=tmp_path,
            # filetype="word",
            format="langchain",
            debug=False,
            verbose=False,
        )

        # Verify output
        assert isinstance(docs, list)
        assert len(docs) > 0
        assert all(isinstance(d, Document) for d in docs)
        # Check that we got some actual content
        assert any(len(d.page_content.strip()) > 0 for d in docs)

    finally:
        # Cleanup
        os.unlink(tmp_path)


@pytest.mark.basic
def test_parse_nytimes():
    """Test parsing the NYTimes homepage."""
    docs = wdoc.parse_file(
        path="https://www.nytimes.com/",
        filetype="auto",
        format="langchain",
        debug=False,
        verbose=False,
    )
    assert isinstance(docs, list)
    assert len(docs) > 0
    assert all(isinstance(d, Document) for d in docs)
    # Check that we got some actual content
    assert any(len(d.page_content.strip()) > 0 for d in docs)


@pytest.mark.experimental
@pytest.mark.skipif(
    " -m experimental" not in " ".join(sys.argv),
    reason="Skip tests of experimental feature by default, use '-m experimental' to run them.",
)
def test_wdoc_env_var_refresh():
    """Test that wdoc env variables are indeed dynamically refreshed."""
    val = copy(WDOC_TYPECHECKING._value)
    assert WDOC_TYPECHECKING == val
    os.environ["WDOC_TYPECHECKING"] = str(val) + "newvalue"
    assert WDOC_TYPECHECKING != val
    assert WDOC_TYPECHECKING == str(val) + "newvalue"
    os.environ["WDOC_TYPECHECKING"] = str(val)
    assert WDOC_TYPECHECKING == val



"""
Chain (logic) used to query a document.
"""

import re
import time

import numpy as np
import pandas as pd
import scipy
import sklearn.decomposition as decomposition
import sklearn.metrics as metrics
import sklearn.preprocessing as preprocessing
from beartype.typing import List, Literal, Tuple, Union
from langchain.docstore.document import Document
from langchain.embeddings import CacheBackedEmbeddings
from langchain_community.chat_models import ChatLiteLLM
from langchain_community.chat_models.fake import FakeListChatModel
from langchain_core.runnables import chain
from langchain_core.runnables.base import RunnableLambda
from langchain_openai import ChatOpenAI
from numpy.typing import NDArray
from tqdm import tqdm

from ..env import WDOC_SEMANTIC_BATCH_MAX_TOKEN_SIZE
from ..errors import (
    InvalidDocEvaluationByLLMEval,
    NoDocumentsAfterLLMEvalFiltering,
    NoDocumentsRetrieved,
)
from ..flags import is_verbose
from ..logger import red, whi
from ..misc import get_tkn_length, thinking_answer_parser
from ..typechecker import optional_typecheck

irrelevant_regex = re.compile(r"\bIRRELEVANT\b")


@optional_typecheck
def check_intermediate_answer(ans: str) -> bool:
    "filters out the intermediate answers that are deemed irrelevant."
    if "<answer>IRRELEVANT</answer>" in ans:
        return False
    if ((not irrelevant_regex.search(ans)) and len(ans) < len("IRRELEVANT") * 2) or len(
        ans
    ) >= len("IRRELEVANT") * 2:
        return True
    return False


@optional_typecheck
def sieve_documents(instance) -> RunnableLambda:
    """cap the number of retrieved documents as if multiple retrievers are used
    we can end up with a lot more document!
    """

    @chain
    @optional_typecheck
    def _sieve(inputs: dict) -> dict:
        assert "question_to_answer" in inputs, inputs.keys()
        assert "unfiltered_docs" in inputs, inputs.keys()
        # we have to pass an instance otherwise we can't know if the top_k got updated
        assert hasattr(instance, "top_k")
        assert hasattr(instance, "max_top_k")
        if instance.max_top_k:
            assert instance.max_top_k >= instance.top_k
        if len(inputs) > instance.top_k:
            red(
                "Number of documents found via embeddings was "
                f"'{inputs['unfiltered_docs']}' which is > top_k ({instance.top_k}) "
                "so we crop"
            )
        inputs["unfiltered_docs"] = inputs["unfiltered_docs"][: instance.top_k]
        return inputs

    return _sieve


@chain
@optional_typecheck
def refilter_docs(inputs: dict) -> List[Document]:
    "filter documents fond via RAG based on the digit answered by the eval llm"
    unfiltered_docs = inputs["unfiltered_docs"]
    evaluations = inputs["evaluations"]
    assert isinstance(
        unfiltered_docs, list
    ), f"unfiltered_docs should be a list, not {type(unfiltered_docs)}"
    assert isinstance(
        evaluations, list
    ), f"evaluations should be a list, not {type(evaluations)}"
    assert len(unfiltered_docs) == len(
        evaluations
    ), f"len of unfiltered_docs is {len(unfiltered_docs)} but len of evaluations is {len(evaluations)}"
    if not unfiltered_docs:
        raise NoDocumentsRetrieved("No document corresponding to the query")

    filtered_docs = []
    for ie, evals in enumerate(evaluations):  # iterating over each document
        if not isinstance(evals, list):
            evals = [evals]
        answers = [thinking_answer_parser(ev)["answer"] for ev in evals]
        for ia, a in enumerate(answers):
            try:
                a = int(a)
            except Exception as err:
                red(
                    f"Document was not evaluated with a number: '{err}' for answer '{a}'\nKeeping the document anyway."
                )
                a = 5
            answers[ia] = a

        if sum(answers) / len(answers) >= 3:
            filtered_docs.append(unfiltered_docs[ie])

    if not filtered_docs:
        raise NoDocumentsAfterLLMEvalFiltering(
            "No document remained after filtering with the query"
        )
    return filtered_docs


@optional_typecheck
def parse_eval_output(output: str) -> str:
    mess = (
        f"The eval LLM returned an output that can't be parsed as expected: '{output}'"
    )
    # empty
    if not output.strip():
        raise InvalidDocEvaluationByLLMEval(mess)

    parsed = thinking_answer_parser(output)

    if is_verbose:
        whi(f"Eval LLM output: '{output}'")

    answer = parsed["answer"]
    try:
        answer = int(answer)
        return str(answer)
    except Exception as err:
        red(
            f"Document was not evaluated with a number: '{err}' for answer '{answer}'\nKeeping the document anyway."
        )
        return str(5)

    if "-" in parsed["answer"]:
        raise InvalidDocEvaluationByLLMEval(mess)
    digits = [d for d in list(parsed["answer"]) if d.isdigit()]

    # contain no digits
    if not digits:
        raise InvalidDocEvaluationByLLMEval(mess)

    # good
    elif len(digits) == 1:
        if digits[0] == "0":
            return "0"
        elif digits[0] == "1":
            return "1"
        elif digits[0] == "2":
            return "1"
        else:
            raise InvalidDocEvaluationByLLMEval(mess)
    else:
        # ambiguous
        raise InvalidDocEvaluationByLLMEval(mess)


@optional_typecheck
def collate_intermediate_answers(
    list_ia: List[str],
) -> str:
    """write the intermediate answers in a single string to be
    combined by the LLM"""
    # remove answers deemed irrelevant
    list_ia = [ia for ia in list_ia if check_intermediate_answer(ia)]
    assert (
        len(list_ia) >= 2
    ), f"Cannot collate a single intermediate answer!\n{list_ia[0]}"

    out = "Intermediate answers:"
    for iia, ia in enumerate(list_ia):
        ia = ia.replace("- • ", "- ").replace("• ", "- ")  # occasional bad md
        out += f"""
<ia source_id={iia + 1}>
{ia}
</ia>\n""".lstrip()
    return out


@optional_typecheck
def semantic_batching(
    texts: List[str],
    embedding_engine: CacheBackedEmbeddings,
) -> List[List[str]]:
    """
    Given a list of text, embed them, do a hierarchical clutering then
    sort the list according to the leaf order, then create buckets that best
    contain each subtopic while keeping a reasonnable number of tokens.
    This probably helps the LLM to combine the intermediate answers
    into one.
    Note that the documents are also sorted inside each batch, so that iterating
    over each document of each batch in order will follow the optimal leaf order.
    """
    max_token = WDOC_SEMANTIC_BATCH_MAX_TOKEN_SIZE

    assert texts, "No input text received"
    assert len(texts) > 1, f"received only one text: {texts}"

    # deduplicate texts
    temp = []
    [temp.append(t) for t in texts if t not in temp]
    texts = temp

    if len(texts) <= 3:
        return [texts]

    text_sizes = {t: get_tkn_length(t) for t in texts}

    # get embeddings
    n_trial = 3
    for trial in range(n_trial):
        try:
            embeds = np.array(embedding_engine.embed_documents(texts)).squeeze()
            break
        except Exception as e:
            red(
                f"Error at trial {trial+1}/{n_trial} when trying to embed texts for semantic batching: '{e}'"
            )
            if trial + 1 >= n_trial:
                red("Too many errors so crashing")
                raise
            else:
                time.sleep(2)

    n_dim = embeds.shape[1]
    assert (
        n_dim > 2
    ), f"Unexpected number of dimension: {n_dim}, shape was {embeds.shape}"

    max_n_dim = min(100, len(texts))

    # optional dimension reduction to gain time
    try:
        if n_dim > max_n_dim:
            scaler = preprocessing.StandardScaler()
            embed_scaled = scaler.fit_transform(embeds)
            pca = decomposition.PCA(n_components=max_n_dim)
            embeds_reduced = pca.fit_transform(embed_scaled)
            assert embeds_reduced.shape[0] == embeds.shape[0]
            vr = np.cumsum(pca.explained_variance_ratio_)[-1]
            if vr <= 0.90:
                red(f"Found lower than exepcted PCA explained variance ratio: {vr:.4f}")
            assert (
                vr >= 0.75
            ), f"Found substancially low explained variance ratio afer pca at {vr:.4f} so not using dimension reduction"
            embeddings = pd.DataFrame(
                columns=[f"v_{i}" for i in range(embeds_reduced.shape[1])],
                index=[i for i in range(len(texts))],
                data=embeds_reduced,
            )
    except Exception as err:
        red(
            f"Error when doing dimension reduction for semantic batching. Original shape: {embeds.shape}. Error: '{err}'\nContinuing anyway."
        )

    if "embeddings" not in locals():
        embeddings = pd.DataFrame(
            columns=[f"v_{i}" for i in range(embeds.shape[1])],
            index=[i for i in range(len(texts))],
            data=embeds,
        )

    # get the pairwise distance matrix
    pairwise_distances = metrics.pairwise_distances
    pd_dist = pd.DataFrame(
        columns=embeddings.index,
        index=embeddings.index,
        data=pairwise_distances(
            embeddings.values,
            n_jobs=-1,
            metric="euclidean",
        ),
    )
    # make sure the intersection is 0 and not a very small float
    for ind in pd_dist.index:
        pd_dist.at[ind, ind] = 0
    # make sure it's symetric
    pd_dist = pd_dist.add(pd_dist.T).div(2)

    # get the hierarchichal semantic sorting order
    dist: NDArray[int] = scipy.spatial.distance.squareform(
        pd_dist.values
    )  # convert to condensed format
    Z: NDArray[Tuple[int, Literal[4]]] = scipy.cluster.hierarchy.linkage(
        dist, method="ward", optimal_ordering=True
    )

    order: NDArray[int] = scipy.cluster.hierarchy.leaves_list(Z)

    # TODO:; if <= 6 texts we should make 2 or 3 batch just using the order

    # # this would just return the list of strings in the best order
    # out_texts = [texts[o] for o in order]
    # assert len(set(out_texts)) == len(out_texts), "duplicates"
    # assert len(out_texts) == len(texts), "extra out_texts"
    # assert not any(o for o in out_texts if o not in texts)
    # assert not any(t for t in texts if t not in out_texts)
    # # whi(f"Done in {int(time.time()-start)}s")
    # assert len(texts) == len(out_texts)

    # get each bucket if we were only looking at the number of texts
    cluster_trials = {}
    cluster_mean_tkn = {}
    for divider in [3, 4, 5, 6]:
        cluster_labels = scipy.cluster.hierarchy.fcluster(
            Z, len(pd_dist.index) // divider, criterion="maxclust"
        )
        labels = np.unique(cluster_labels)
        labels.sort()
        if len(labels) == 1:  # re cluster if only one label found
            continue

        # use heuristics to find the best number of dividers by looking
        # at the average number of token in each clusters
        total_mean = 0
        for lab in labels:
            lt = [texts[int(ind)] for ind in np.argwhere(cluster_labels == lab)]
            lsize = sum([text_sizes[t] for t in lt])
            lmean = lsize / len(lt)
            total_mean += lmean
        total_mean /= len(labels)
        cluster_mean_tkn[divider] = total_mean
        cluster_trials[divider] = cluster_labels

    best_clusters = None
    for d, ct in cluster_mean_tkn.items():
        if ct < max_token and ct >= max_token / 2:
            best_clusters = cluster_trials[d]
            break
    if best_clusters is None:
        best_tkns = min(list(cluster_mean_tkn.values()))
        for d, ct in cluster_mean_tkn.items():
            if ct == best_tkns:
                best_clusters = cluster_trials[d]
                break
    assert best_clusters is not None
    cluster_labels = best_clusters

    labels = np.unique(cluster_labels)
    labels.sort()

    assert len(labels) > 1, cluster_labels

    # make sure no cluster contains only one text
    while not all((cluster_labels == lab).sum() > 1 for lab in labels):
        if is_verbose:
            whi("Remapping clusters.")
        for lab in labels:
            if (cluster_labels == lab).sum() == 1:
                t = texts[np.argmax(cluster_labels == lab)]
                # the closest is always itself so checking the 2nd closest
                t_close = (pd_dist.loc[texts.index(t), :]).nsmallest(2).index.tolist()
                assert texts.index(t) == t_close[0]
                t_closest = t_close[1]
                l_closest = cluster_labels[t_closest]
                if (cluster_labels == l_closest).sum() + 1 == len(texts):
                    # merging small to big would result in only one cluster:
                    # better to even them out
                    assert len(labels) == 2, labels
                    cluster_labels[t_closest] = lab
                    if is_verbose:
                        whi(f"Remapped one item from cluster {l_closest} to {lab}")
                else:  # good to go
                    cluster_labels[cluster_labels == lab] = l_closest
                    if is_verbose:
                        whi(f"Remapped single item of cluster {lab} to {l_closest}")
                break
        labels = np.unique(cluster_labels)
        labels.sort()
    assert len(labels) > 1, cluster_labels
    assert all((cluster_labels == lab).sum() > 1 for lab in labels), cluster_labels

    # Create buckets
    buckets = []
    current_bucket = []
    current_tokens = 0

    # fill each bucket until reaching max_token
    for lab in labels:
        lab_ind = np.argwhere(cluster_labels == lab)
        assert len(lab_ind) > 1, f"{lab_ind}\n{cluster_labels}"
        assert len(lab_ind) < len(texts), f"{lab_ind}\n{cluster_labels}"
        for clustid in lab_ind:
            text = texts[int(clustid)]
            size = text_sizes[text]
            if (current_tokens + size > max_token) and current_bucket:
                buckets.append(current_bucket)
                current_bucket = [text]
                current_tokens = 0
            else:
                current_bucket.append(text)
                current_tokens += size

        assert current_bucket
        buckets.append(current_bucket)
        current_bucket = []
        current_tokens = 0
    assert all(bucket for bucket in buckets), "Empty buckets"

    # sort each bucket based on the optimal order
    for ib, b in enumerate(buckets):
        buckets[ib] = sorted(b, key=lambda t: order[texts.index(t)])

    # now if any bucket contains only one text, that means it has too many
    # tokens itself, so we reequilibrate from the previous buckets
    while not all(len(b) >= 2 for b in buckets):
        if is_verbose:
            whi(f"Merging sub buckets. Current len: {len(buckets)}")
        for ib, b in enumerate(buckets):
            assert b
            if len(b) == 1:
                # figure out which bucket to merge with
                if ib == 0:  # first , merge with next
                    next_id = ib + 1
                elif ib + 1 == len(buckets):  # last, take the penultimate
                    next_id = ib - 1
                elif ib != len(
                    buckets
                ):  # not first nor last, take the neighbour with least minimal distance
                    t_cur = b[0]
                    prev = min(
                        [
                            pd_dist.loc[texts.index(t_cur), texts.index(t)]
                            for t in buckets[ib - 1]
                        ]
                    )
                    next = min(
                        [
                            pd_dist.loc[texts.index(t_cur), texts.index(t)]
                            for t in buckets[ib + 1]
                        ]
                    )
                    assert prev > 0 and next > 0
                    if prev < next:
                        next_id = ib - 1
                    else:
                        next_id = ib + 1
                assert buckets[next_id], buckets[next_id]
                if is_verbose:
                    whi(f"Next_id is {next_id}")

                if len(buckets[next_id]) == 1:  # both texts are big, merge them anyway
                    if next_id > ib:
                        buckets[next_id].insert(0, b.pop())
                    else:
                        buckets[next_id].append(b.pop())
                    assert not b, b
                elif (
                    len(buckets[next_id]) == 2
                ):  # merging 2:1 -> 1:2 would create a loop
                    if next_id > ib:
                        buckets[next_id].insert(0, b.pop())
                    else:
                        buckets[next_id].append(b.pop())
                    assert not b, b
                else:
                    # send text to the next bucket, at the correct position
                    if next_id > ib:
                        b.append(buckets[next_id].pop(0))
                    else:
                        b.append(buckets[next_id].pop(-1))
                assert id(b) == id(buckets[ib])
                break
        buckets = [b for b in buckets if b]
    assert all(
        len(b) >= 2 for b in buckets
    ), f"Invalid size of buckets: '{[len(b) for b in buckets]}'"

    unchained = []
    [unchained.extend(b) for b in buckets]
    assert len(unchained) == len(
        set(unchained)
    ), "There were duplicate texts in buckets!"
    assert all(t in texts for t in unchained), "Some text of buckets were added!"

    return buckets


@optional_typecheck
def pbar_chain(
    llm: Union[ChatLiteLLM, ChatOpenAI, FakeListChatModel],
    len_func: str,
    **tqdm_kwargs,
) -> RunnableLambda:
    "create a chain that just sets a tqdm progress bar"

    @chain
    def actual_pbar_chain(
        inputs: Union[dict, List],
        llm: Union[ChatLiteLLM, ChatOpenAI, FakeListChatModel] = llm,
    ) -> Union[dict, List]:

        llm.callbacks[0].pbar.append(
            tqdm(
                total=eval(len_func),
                **tqdm_kwargs,
            )
        )
        if not llm.callbacks[0].pbar[-1].total:
            red(f"Empty total for pbar: {llm.callbacks[0].pbar[-1]}")

        return inputs

    return actual_pbar_chain


@optional_typecheck
def pbar_closer(
    llm: Union[ChatLiteLLM, ChatOpenAI, FakeListChatModel],
) -> RunnableLambda:
    "close a pbar created by pbar_chain"

    @chain
    def actual_pbar_closer(
        inputs: Union[dict, List],
        llm: Union[ChatLiteLLM, ChatOpenAI, FakeListChatModel] = llm,
    ) -> Union[dict, List]:
        pbar = llm.callbacks[0].pbar[-1]
        pbar.update(pbar.total - pbar.n)
        pbar.n = pbar.total
        pbar.close()

        return inputs

    return actual_pbar_closer



"""
Prompts used by wdoc.
"""

from beartype.typing import List
from langchain_core.messages import HumanMessage, SystemMessage
from langchain_core.output_parsers import PydanticOutputParser
from langchain_core.prompts import ChatPromptTemplate, PromptTemplate
from langchain_core.prompts.chat import (
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)
from pydantic import BaseModel, Field, model_validator

from .logger import red, whi
from .misc import get_tkn_length

# PROMPT FOR SUMMARY TASKS
BASE_SUMMARY_PROMPT = ChatPromptTemplate(
    messages=[
        SystemMessagePromptTemplate.from_template(
            """
You are Sam, the best employee of WDOC-CORP©, the best of my team. Your goal today is to summarize in a specific way a text section I just sent you, but I'm not only interested in high level takeaways. I also need the thought process present in the document, the reasoning followed, the arguments used etc. But your summary has to be as quick and easy to read as possible while following specific instructions.
This is very important to me so if you succeed, I'll pay you up to $2000 depending on how well you did!

<instructions>
- In some cases, I can give you additional instructions, you have to treat them as the present rules.
- Take a deep breath before answering
- Include:
    - All noteworthy information, anecdotes, facts, insights, definitions, clarifications, explanations, ideas, technical details, etc
    - Epistemic indicators: you need to make explicit what markers of uncertainty for each information
- Exclude:
    - Sponsors, advertisements, etc
    - Jokes, ramblings
    - End of page references and sources, tables of content, links etc
    - When in doubt about whether to include an information, include it
- Format:
    - Use markdown format: that means logical indentation, bullet points, bold etc
        - DON'T use markdown headers, only bullet points.
        - DON'T: use a single bullet point for the whole answer. Eg if the question is about cancer, instead of 1 top level 'cancer' bullet, prefer several top level bullets for each aspect of the answer.
        - Use bold for important concepts, and italic for epistemic markers
            - ie "- *In his opinion*, **dietary supplements** are **healthy** because ..."
        - Use as many indentation levels as necessary to respect the rules. I don't mind if there are even 10 levels!
    - Stay faithful to the tone of the author
    - You don't always have to use full sentences: you can ignore end of line punctuation etc
        - BUT it is more important to be unambiguous and truthful than concise
        - EVERY TIME POSSIBLE: use direct quote, 'formatted like that'
    - Use one bullet point per information
        - With the use of logical indentation this makes the whole piece quick and easy to skim
    - Write your summary in {language}
    - Avoid repetitions
        - eg don't start several bullet points by 'The author thinks that', just say it once then use indented children bullet points to make it implicit
    - If the text includes timestamps, write the approximate time of the sections in the top level bullet points.
</instructions>
""".strip()
        ),
        HumanMessagePromptTemplate.from_template(
            """
{recursion_instruction}{metadata}{previous_summary}

<text_section>
{text}
</text_section>
""".strip()
        ),
    ],
    input_variables=[
        "language",
        "recursion_instruction",
        "metadata",
        "previous_summary",
        "text",
    ],
    validate_template=True,
)

# if the text to summarize is long, give the end of the previous summary to help with transitions
PREV_SUMMARY_TEMPLATE = """

<end_of_summary_of_prev_section>
{previous_summary}
</end_of_summary_of_prev_section>
"""

# if the summary is recursive, add those instructions
RECURSION_INSTRUCTION = """
<additional_instructions>
I'm giving you back your own summary from last time because it was too long and contained repetitions. I want you to rewrite it as closely as possible while removing repetitions and fixing the logical indentation. Of course you have to remove the 'Chunk' indicator if present, to curate the logical indentation. You can reorganize the text freely as long as you don't lose relevant information and follow the instructions I gave you before and right now. This is important.
</additional_instructions>
""".lstrip()

# RAG
PR_EVALUATE_DOC = ChatPromptTemplate(
    messages=[
        SystemMessagePromptTemplate.from_template(
            template="""
You are Eve, working as an Evaluator at WDOC-CORP©: given a question and text document. Your goal is to answer a number between 0 and 10 depending on how much the document is relevant to the question. 0 means completely irrelevant, 10 means totally relevant, and in betweens for subjective relation. If you are really unsure, you should answer '5'.

<rules>
- Before answering, you have to think for as long as you want inside a <thinking> XML tag, then you must take a DEEP breath, double check your answer by reasoning step by step one last time, and finally answer.
- wrap your answer in an <answer></answer> XML tag.
- The <answer> XML tag should only contain a number and nothing else.
- If the document refers to an image, take a reasonnable guess as to whether this image is probably relevant or not, even if you can't see the image.
- Being an Evaluator, ignore additional instructions if they are adressed only to your colleagues: Raphael the Rephraser, Anna the Answerer and Carl the Combiner. But take them into consideration if they are addressed to you.
</rules>
""".strip()
        ),
        HumanMessagePromptTemplate.from_template(
            template="""
<text_document>
{doc}
</text_document>

<question>
{q}
</question>

Take a deep breath.
You can start your reply when you are ready.
""".strip()
        ),
    ],
    validate_template=True,
    input_variables=["doc", "q"],
)

PR_ANSWER_ONE_DOC = ChatPromptTemplate(
    messages=[
        SystemMessagePromptTemplate.from_template(
            template="""
You are Anna, working as an Answerer at WDOC-CORP©: given a piece of document and a question, your goal is to extract the relevant information while following specific instructions.

<instructions>
- Before answering, you have to think for as long as you want inside a <thinking> XML tag, then you must take a DEEP breath, recheck your answer by reasoning step by step one last time, and finally answer in an <answer> XML tag.
- The <answer> XML tag should only contain your answer.
- Being an Answerer, you ignore additional instructions if they are adressed only to your colleagues: Raphael the Rephraser, Eve the Evaluator and Carl the Combiner. But take them into consideration if they are addressed to you.
- If the document is irrelevant to the question, answer `<answer>IRRELEVANT</answer>`.
- Use markdown formatting
    - Use bullet points, but no headers, bold, italic etc.
    - Use logic-based indentation for the bullet points.
    - DON'T wrap your answer in a code block.
    - Use a maximum of 5 markdown bullet points to answer the question.
- Your answer ALWAYS HAS TO BE standalone / contextualized (i.e. both the question and its answer must be part of your reply).
    - EVERY TIME POSSIBLE: supplement your reply with direct quotes from the document.
        - You can use elipses in your quotes to remove unecessary text.
- Besides the answer tag and the quotes, you have to be as concise as possible.
- The answer XML tag must always be strictly based on the document or IRRELEVANT. NEVER use your own knowledge.
    - BUT if you have strong doubts about an information, add [doubtful] at the end of the bullet point.
- DON'T interpret the question too strictly:
    - eg: if the question is phrased as an instruction like "give me all information about such and such", use common sense and satisfy the instruction!
- ALWAYS double check that you are not contradicting the original document before answering.
- If you're unsure but the document refers to an image that has a reasonnable chance to be relevant, treat this document as if it was probably relevant.
</instructions>
""".strip()
        ),
        HumanMessagePromptTemplate.from_template(
            template="""
<context>
{context}
</context>

<question>
{question_to_answer}
</question>

Now take a deep breath.
Take your time.
Start your reply when you're ready.
""".strip()
        ),
    ],
    validate_template=True,
    input_variables=["context", "question_to_answer"],
)

PR_COMBINE_INTERMEDIATE_ANSWERS = ChatPromptTemplate(
    messages=[
        SystemMessagePromptTemplate.from_template(
            template="""
You are Carl, working as a Combiner at WDOC-CORP©: given a question and partial answers, your goal is to:
- combine all partial answers to answer the original question as md bullet points,
- while combining all additional information as additional bullet points.
- And keeping track of sources.

<instructions>
- Being a Combiner, ignore additional instructions if they are adressed only to your colleagues: Raphael the Rephraser, Eve the Evaluator and Anna the Answerer. But take them into consideration if they are addressed to you.
- Format:
    - Use markdown format, with bullet points.
      - IMPORTANT: use logical indentation to organize information hierarchically.
        - Use as many indentation levels as necessary to respect the rules. I don't mind if there are even 10 levels!
      - The present instructions are a good example of proper formatting.
    - You are allowed to use markdown tables if appropriate.
    - Reuse acronyms without specifying what they mean.
    - Be concise but don't omit only irrelevant information from the statements.
    - Answer in the same language as the question.
- What to include:
    - If an answer refers to an image that has a reasonnable chance of being relevant, treat that as possibly relevant information.
    - Ignore statements that are completely irrelevant to the question.
    - Semi relevant statements can be included, especially if related to possible followup questions.
    - No redundant information must remain.
        - eg: fix redundancies with one parent bullet point and several indented children.
    - DON'T interpret the question too strictly:
        - eg: if the question makes reference to "documents" consider that it's what I call here "statements" for example.
        - eg: if the question is phrased as an instruction like "give me all information about such and such", use common sense and satisfy the instruction!
- The iia can consist of a succession of thoughts in <thinking> XML tag followd by the answer in an <answer> XML tag. In that case you have to only take into account the <answer> (the <thinking> can still be helpful but don't treat it as source you can include in your own answer, you can't!)
- If some information are imcompatible, don't make a judgement call: just include both and add short clarification between parentheses and I'll take a look.
- Sources are designated by unique source_id. Use the format [id1, id2], to keep track of each source so that we can find the original source of each information in your final answer.
    - Ideally, the sources are mentionned as close as possible to the key information, and always at the end of the bullet point.
    - It is extremely important that you do not forget to include a source.
    - Note that a previous employee of wdoc can in some situation add a source [doubtful] to indicate that he's suspicious of a specific information. Keep track of this as if it were a source.
- Only use information from the provided statements.
    - IMPORTANT: if the statements are insufficient to answer the question you MUST specify as source [OPINION] (i.e. 'OPINION' is a valid source id). This is important, it allows me to know that the source was you for a specific information.
- Before answering, you have to think for as long as you want inside a <thinking> XML tag, then you must take a DEEP breath, recheck your answer by reasoning step by step one last time, and finally answer.
- wrap your answer in an <answer> XML tag.
- The <answer> XML tag should only contain your answer.
</instructions>
""".strip()
        ),
        HumanMessagePromptTemplate.from_template(
            template="""
<question>
{question}
</question>

<partial_answers>
{intermediate_answers}
</partial_answers>

Now take a deep breath.
Take your time.
Start your reply when you're ready.
""".strip()
        ),
    ],
    validate_template=True,
    input_variables=["question", "intermediate_answers"],
)

# embedding queries


# https://python.langchain.com/docs/how_to/output_parser_structured/
class ExpandedQuery(BaseModel):
    thoughts: str = Field(description="Reasoning you followed")
    output_queries: List[str] = Field(description="List containing each output query")

    @model_validator(mode="before")
    @classmethod
    def nonempty_queries(cls, values: dict) -> dict:
        oq = values["output_queries"]
        if not isinstance(oq, list):
            raise ValueError("output_queries has to be a list")
        if not oq:
            raise ValueError("output_queries can't be empty")
        if not all(isinstance(q, str) for q in oq):
            raise ValueError("output_queries has to be a list of str")
        oq = [q.strip() for q in oq if q.strip()]
        if not oq:
            raise ValueError(
                "output_queries can't be empty after removing empty strings"
            )
        return values

    @model_validator(mode="after")
    @classmethod
    def remove_thoughts(cls, values) -> List[str]:
        return values.output_queries


multiquery_parser = PydanticOutputParser(pydantic_object=ExpandedQuery)
PR_MULTI_QUERY_PROMPT = ChatPromptTemplate(
    messages=[
        # has to be a systemmessage directly otherwise pydantic's fstrings impact the validation
        SystemMessage(
            content="""
You are Raphael, working as a Rephraser at WDOC-CORP©: given a user
RAG query, your task is to expand the query (i.e. you generate 10 different
versions of the query to increase the chances of retrieving the most
relevant documents using a vector embedding search.
You can generate multiple perspectives as well as anticipate the actual
answer (like HyDE style search).
You will be given specific formatting your for your answer in due time.
I know this is a complex task so you are encouraged to express your
thinking process BEFORE answering (as long as you respect the format).

Example: if you are given a short query "breast cancer", you should expand
the query to a list of 10 similar query like "breast cancer treatment",
"diagnostics of breast cancer", "epidemiology of breast cancer", "clinical
presentation of breast cancers", "classification of breast cancer", "evaluation
of breast cancer", etc.
You can also anticipate the answer like "the most used chemotherapies
for breast cancers are anthracyclines, taxanes and cyclophosphamide".

DO NOT reply anything that is not either a thought or the queries. You
have to do your thinking INSIDE the thought json.
Also, being a Rephraser, you can take into consideration additional
instructions that are adressed to your colleagues: Eve the Evaluator, Anan the
Answerer and Carl the Combiner. They can sometimes be useful to expand the queries.
""".strip()
            + "\n"
            + multiquery_parser.get_format_instructions()
        ),
        HumanMessagePromptTemplate.from_template(
            template="Here's the query to expand: '''{question}'''" ""
        ),
    ],
    input_variables=["question"],
    validate_template=True,
)


class Prompts_class:
    evaluate: ChatPromptTemplate
    answer: ChatPromptTemplate
    combine: ChatPromptTemplate
    multiquery: PromptTemplate

    def __init__(
        self,
        **prompts: dict[str, ChatPromptTemplate],
    ) -> None:
        for key, pr in prompts.items():
            setattr(self, key, pr)

    def enable_prompt_caching(self, prompt_key: str) -> None:
        assert prompt_key in [
            "evaluate",
            "answer",
            "combine",
            "multiquery",
        ], "Unexpected prompt_key"

        whi(
            f"Detected anthropic in llm name so enabling anthropic prompt_caching for {prompt_key} prompt"
        )
        prompt = getattr(self, prompt_key)
        sys = prompt.messages[0]
        assert isinstance(sys, (SystemMessagePromptTemplate, SystemMessage)), sys
        if hasattr(sys, "content"):
            content = sys.content
        else:
            content = sys.prompt.template
        #
        # anthropic caching works only above 1024 tokens but doesn't count
        # exactly like openai
        tkl = get_tkn_length(content)
        if tkl < 750:
            red(
                f"System prompt is only {tkl} openai tokens so caching will won't work, not using it."
            )
            return

        new_content = [
            {"type": "text", "text": content, "cache_control": {"type": "ephemeral"}}
        ]
        if hasattr(sys, "content"):
            prompt.messages[0].content = new_content
        else:
            prompt.messages[0].prompt.template = new_content


prompts = Prompts_class(
    evaluate=PR_EVALUATE_DOC,
    answer=PR_ANSWER_ONE_DOC,
    combine=PR_COMBINE_INTERMEDIATE_ANSWERS,
    multiquery=PR_MULTI_QUERY_PROMPT,
)



"""
Custom embeddings to use litellm. This allows using for example
"ollama/bge-m3" as a model name.
Source: https://python.langchain.com/docs/how_to/custom_embeddings/
"""

from typing import List, Optional

import litellm
from langchain_core.embeddings import Embeddings


class LiteLLMEmbeddings(Embeddings):
    """Litellm embedding model integration."""

    def __init__(
        self,
        model: str,
        dimensions: Optional[int],
        api_base: Optional[str],
        private: bool,
        **embed_kwargs,
    ):
        assert (
            "/" in model
        ), "model must contain a /, for example 'ollama/bge-m3' or 'openai/text-embedding-ada-002'"
        if private:
            if not api_base:
                assert any(
                    provider in model for provider in ["ollama", "huggingface"]
                ), "--private argument is set and api_base not overridden BUT the model does not contain ollama nor huggingface, this can be a mistake so crashing out of abundance of caution. If you think this is a bug please open an issue on github."
        self.model = model
        self.dimensions = dimensions
        self.private = private
        self.api_base = api_base
        self.embed_kwargs = embed_kwargs

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed search docs."""
        # https://docs.litellm.ai/docs/embedding/supported_embedding
        vecs = litellm.embedding(
            model=self.model,
            input=texts,
            dimensions=self.dimensions,
            encoding_format="float",
            timeout=600,
            api_base=self.api_base,
            user="wdoc_embeddings",
            drop_params=True,
            # 'sentence-similarity', 'feature-extraction', 'rerank', 'embed', 'similarity'
            # input_type="feature-extraction",  # seems to crash for openai
            **self.embed_kwargs,
        )
        if hasattr(
            vecs, "data"
        ):  # must an EmbeddingsResponse format, for example ollama
            data = vecs.data
            if isinstance(data, list) and isinstance(data[0], dict):
                vecs = [v["embedding"] for v in data]
            elif isinstance(data, list) and hasattr(data[0], "embedding"):
                vecs = [v.embedding for v in data]
            else:
                raise Exception(
                    f"Failed to parsed output of litellm embedding for model '{self.model}'. String rendering is '{vecs}'. Please open a github issue to get that fixed."
                )
        return vecs

    def embed_query(self, text: str) -> List[float]:
        """Embed query text."""
        return self.embed_documents([text])[0]



"""
* Class used to create the embeddings.
* Loads and store embeddings for each document.
"""

# import math
import hashlib
import os
import random
import time
from functools import wraps
from pathlib import Path

import litellm
import numpy as np
from beartype.typing import Any, Callable, List, Optional, Tuple, Union
from joblib import Parallel, delayed
from langchain.embeddings import CacheBackedEmbeddings
from langchain_community.embeddings import (
    HuggingFaceEmbeddings,
    HuggingFaceInstructEmbeddings,
    SentenceTransformerEmbeddings,
)
from langchain_core.vectorstores.base import VectorStore
from langchain_community.vectorstores import FAISS
from langchain_core.embeddings import Embeddings
from langchain_openai import OpenAIEmbeddings
from tqdm import tqdm

# from langchain.storage import LocalFileStore
from .customs.compressed_embeddings_cacher import LocalFileStore
from .customs.litellm_embeddings import LiteLLMEmbeddings
from .env import (
    WDOC_DEFAULT_EMBED_DIMENSION,
    WDOC_EXPIRE_CACHE_DAYS,
    WDOC_MOD_FAISS_SCORE_FN,
)
from .flags import is_verbose
from .logger import red, whi
from .misc import ModelName, cache_dir, get_tkn_length
from .typechecker import optional_typecheck


def status(message: str):
    if is_verbose:
        whi(f"STATUS: {message}")


(cache_dir / "faiss_embeddings").mkdir(exist_ok=True)

# Source: https://api.python.langchain.com/en/latest/_modules/langchain_community/embeddings/huggingface.html#HuggingFaceEmbeddings
DEFAULT_EMBED_INSTRUCTION = "Represent the document for retrieval: "
DEFAULT_QUERY_INSTRUCTION = (
    "Represent the question for retrieving supporting documents: "
)


if WDOC_MOD_FAISS_SCORE_FN:

    def score_function(distance: float) -> float:
        """
        Scoring function for faiss to make sure it's positive.
        Related issue: https://github.com/langchain-ai/langchain/issues/17333

        In langchain the default value is the euclidean relevance score:
        return 1.0 - distance / math.sqrt(2)

        The output is a similarity score: it must be [0,1] such that
        0 is the most dissimilar, 1 is the most similar document.
        """
        # To disable it but simply check: uncomment this and add "import math"
        # assert distance >= 0, distance
        # return 1.0 - distance / math.sqrt(2)
        new = 1 - ((1 + distance) / 2)
        return new

else:
    score_function = None


@optional_typecheck
def load_embeddings_engine(
    modelname: ModelName,
    cli_kwargs: dict,
    api_base: Optional[str],
    embed_kwargs: dict,
    private: bool,
    do_test: bool,
) -> CacheBackedEmbeddings:
    """
    Create the Embeddings class used to compute embeddings. This class is wrapped
    into a CacheBackedEmbeddings to add a caching layer.
    """
    if "embed_instruct" in cli_kwargs and cli_kwargs["embed_instruct"]:
        instruct = True
    else:
        instruct = False

    if is_verbose:
        whi(f"Selected embedding model '{modelname}' of backend {modelname.backend}")

    if True:
        try:
            embeddings = LiteLLMEmbeddings(
                model=modelname.original,
                dimensions=WDOC_DEFAULT_EMBED_DIMENSION,  # defaults to None
                api_base=api_base,
                private=private,
                **embed_kwargs,
            )
            if do_test:
                test_embeddings(embeddings)
        except Exception as e:
            red(
                f"Failed to use the experimental LiteLLMEmbeddings backend, defaulting to using the previous implementation. Error was '{e}'. Please open a github issue to help the developper debug this until it is stable enough."
            )

    if "embeddings" in locals():
        # already loaded
        pass

    elif modelname.backend == "openai":
        if private:
            assert api_base, "If private is set, api_base must be set too"
        else:
            assert (
                "OPENAI_API_KEY" in os.environ
                and os.environ["OPENAI_API_KEY"]
                and "REDACTED" not in os.environ["OPENAI_API_KEY"]
            ), "Missing OPENAI_API_KEY"

        embeddings = OpenAIEmbeddings(
            model=modelname.model,
            openai_api_key=os.environ["OPENAI_API_KEY"],
            api_base=api_base,
            dimensions=WDOC_DEFAULT_EMBED_DIMENSION,  # defaults to None
            **embed_kwargs,
        )

    elif modelname.backend == "huggingface":
        assert (
            not private
        ), f"Set private but tried to use huggingface embeddings, which might not be as private as using sentencetransformers"
        model_kwargs = {
            "device": "cpu",
            # "device": "cuda",
        }
        model_kwargs.update(embed_kwargs)
        if modelname.backend == "google" and "gemma" in modelname.model.lower():
            assert (
                "HUGGINGFACE_API_KEY" in os.environ
                and os.environ["HUGGINGFACE_API_KEY"]
                and "REDACTED" not in os.environ["HUGGINGFACE_API_KEY"]
            ), "Missing HUGGINGFACE_API_KEY"
            hftkn = os.environ["HUGGINGFACE_API_KEY"]
            # your token to use the models
            model_kwargs["use_auth_token"] = hftkn
        if instruct:
            embeddings = HuggingFaceInstructEmbeddings(
                model_name=modelname.model,
                model_kwargs=model_kwargs,
                embed_instruction=DEFAULT_EMBED_INSTRUCTION,
                query_instruction=DEFAULT_QUERY_INSTRUCTION,
            )
        else:
            embeddings = HuggingFaceEmbeddings(
                model_name=modelname.model,
                model_kwargs=model_kwargs,
            )

        if modelname.backend == "google" and "gemma" in modelname.model.lower():
            # please select a token to use as `pad_token` `(tokenizer.pad_token = tokenizer.eos_token e.g.)`
            # or add a new pad token via `tokenizer.add_special_tokens({'pad_token': '[pad]'})
            embeddings.client.tokenizer.pad_token = (
                embeddings.client.tokenizer.eos_token
            )

    elif modelname.backend == "sentencetransformers":
        if private:
            red(f"Private is set and will use sentencetransformers backend")
        embed_kwargs.update(
            {
                "batch_size": 1,
                "device": None,
            }
        )
        embeddings = SentenceTransformerEmbeddings(
            model_name=modelname.model,
            encode_kwargs=embed_kwargs,
        )

    else:
        raise ValueError(f"Invalid embedding backend: {modelname.backend}")

    if do_test:
        try:
            test_embeddings(embeddings)
        except Exception as e:
            red(
                f"Error when testing embeddings, something is probably wrong with the backend. Error is '{e}'. Please open a github issue to help the developper"
            )

    lfs = LocalFileStore(
        database_path=cache_dir / "CacheEmbeddings" / modelname.sanitized,
        expiration_days=WDOC_EXPIRE_CACHE_DAYS,
        verbose=is_verbose,
    )

    cache_content = list(lfs.yield_keys())
    whi(f"Found {len(cache_content)} embeddings in local cache")

    # cached_embeddings = embeddings
    cached_embeddings = CacheBackedEmbeddings.from_bytes_store(
        embeddings,
        lfs,
        namespace=modelname.sanitized,
    )

    if do_test:
        try:
            test_embeddings(cached_embeddings)
        except Exception as e:
            red(
                f"Error when testing embeddings after loading the cache, something is probably wrong with the backend. Error is '{e}'. Please open a github issue to help the developper"
            )

    return cached_embeddings


@optional_typecheck
def create_embeddings(
    modelname: ModelName,
    cached_embeddings: Embeddings,
    save_embeds_as: Union[str, Path],
    load_embeds_from: Optional[Union[str, Path]],
    loaded_docs: Any,
    dollar_limit: Union[int, float],
    private: bool,
) -> VectorStore:
    """
    For each document of loaded_docs, we check if the embeddings were already
    computed and present in the cache or ask the CacheBackedEmbeddings class
    to create them and return to wdoc.loaded_embeddings.
    """

    # reload passed embeddings
    if load_embeds_from:
        red("Reloading documents and embeddings from file")
        path = Path(load_embeds_from)
        assert path.exists(), f"file not found at '{path}'"
        db = FAISS.load_local(
            str(path),
            cached_embeddings,
            relevance_score_fn=score_function,
            allow_dangerous_deserialization=True,
        )
        n_doc = len(db.index_to_docstore_id.keys())
        red(f"Loaded {n_doc} documents")
        return db, cached_embeddings

    whi("\nLoading embeddings.")

    db = None
    ti = time.time()
    docs = loaded_docs
    whi(f"Docs to embed: {len(docs)}")

    # check price of embedding
    full_tkn = sum([get_tkn_length(doc.page_content) for doc in docs])
    whi(f"Total number of tokens in documents: '{full_tkn}'")
    if modelname.backend in [
        "ollama",
        "huggingface",
        "sentence-transformers",
        "sentencetransformers",
    ]:
        price = 0
        whi("Local embedding model detected, setting the price to 0")
    else:
        if private:
            whi("Not checking token price because private is set")
            price = 0
        elif modelname.original in litellm.model_cost:
            price = litellm.model_cost[modelname.original]["input_cost_per_token"]
            assert litellm.model_cost[modelname.original]["output_cost_per_token"] == 0
        elif modelname.model in litellm.model_cost:
            price = litellm.model_cost[modelname.model]["input_cost_per_token"]
            assert litellm.model_cost[modelname.model]["output_cost_per_token"] == 0
        else:
            red(
                f"Couldn't find the price of embedding model {modelname.original}. Assuming the cost is zero"
            )
            price = 0

    dol_price = full_tkn * price
    red(f"Total cost to embed all tokens is ${dol_price:.6f}")
    if dol_price > dollar_limit:
        ans = input("Do you confirm you are okay to pay this? (y/n)\n>")
        if ans.lower() not in ["y", "yes"]:
            red("Quitting.")
            raise SystemExit()

    # create a faiss index for batch of documents
    ts = time.time()
    batch_size = 1000
    batches = [
        [i * batch_size, (i + 1) * batch_size]
        for i in range(len(docs) // batch_size + 1)
    ]

    @optional_typecheck
    def embed_one_batch(
        batch: List,
        ib: int,
    ) -> VectorStore:
        n_trial = 3
        for trial in range(n_trial):
            # whi(f"Embedding batch #{ib + 1}")
            try:
                temp = FAISS.from_documents(
                    batch,
                    cached_embeddings,
                    normalize_L2=True,
                    relevance_score_fn=score_function,
                )
                break
            except Exception as e:
                red(
                    f"Thread #{ib + 1} Error at trial {trial+1}/{n_trial} when trying to embed documents: {e}"
                )
                if trial + 1 >= n_trial:
                    red("Too many errors: bypassing the cache:")
                    temp = FAISS.from_documents(
                        batch,
                        cached_embeddings.underlying_embeddings,
                        normalize_L2=True,
                        relevance_score_fn=score_function,
                    )
                    break
                else:
                    time.sleep(1)
        return temp

    temp_dbs = Parallel(
        backend="threading",
        n_jobs=10,
        verbose=0 if not is_verbose else 51,
    )(
        delayed(embed_one_batch)(
            batch=docs[batch[0] : batch[1]],
            ib=ib,
        )
        for ib, batch in tqdm(
            enumerate(batches),
            total=len(batches),
            desc="Embedding by batch",
            # disable=not is_verbose,
        )
    )
    for temp in temp_dbs:
        if not db:
            db = temp
        else:
            db.merge_from(temp)

    whi(f"Done creating index (total time: {time.time()-ti:.2f}s)")

    # saving embeddings
    db.save_local(save_embeds_as)

    return db


@optional_typecheck
def test_embeddings(embeddings: Embeddings) -> None:
    "Simple testing of embeddings to know early if something seems wrong"
    vec1 = np.array(embeddings.embed_query("This is a test"))
    vec2 = np.array(embeddings.embed_documents(["This is another test"])[0])
    shape1 = vec1.shape
    shape2 = vec2.shape
    assert (
        shape1 == shape2
    ), f"Test vectors 1 has shape {shape1} but vector 2 has shape {shape2}"
    assert not (
        vec1 == vec2
    ).all(), f"Test vectors 1 and 2 are identical despite different inputs"
    assert not (
        (vec1 == 0).all() or (vec2 == 0).all()
    ), "Test vectors 1 or 2 or both is only zeroes"
